from __future__ import annotations

import itertools
import weakref

import claripy
from claripy import backends, errors, false
from claripy.ast import Base
from claripy.errors import UnsatError


class ModelCache:
    """ModelCache is a wrapper around a model that provides some additional functionality."""

    def __init__(self, model):
        self.model = model
        self.replacements = {}
        self.constraint_only_replacements = {}

    def __hash__(self):
        if not hasattr(self, "_hash"):
            self._hash = hash(frozenset(self.model.items()))  # pylint:disable=attribute-defined-outside-init
        return self._hash

    def __eq__(self, other):
        return self.model == other.model

    def __getstate__(self):
        return (self.model,)

    def __setstate__(self, s):
        self.model = s[0]
        self.replacements = {}
        self.constraint_only_replacements = {}

    #
    # Splitting support
    #

    def filter(self, variables):
        return ModelCache({k: self.model[k] for k in self.model if k in variables})

    @staticmethod
    def combine(*models):
        return ModelCache(dict(itertools.chain.from_iterable(m.model.items() for m in models)))

    #
    # Model-driven evaluation
    #

    def _leaf_op(self, a):
        if a.op == "BVS":
            return claripy.BVV(self.model.get(a.args[0], 0), a.length)
        if a.op == "BoolS":
            return claripy.BoolV(self.model.get(a.args[0], True))
        if a.op == "FPS":
            return claripy.FPV(self.model.get(a.args[0], 0.0), a.args[1])
        if a.op == "StringS":
            return claripy.StringV(self.model.get(a.args[0], ""))
        return a

    def _leaf_op_existonly(self, a):
        if a.op == "BVS":
            return claripy.BVV(self.model[a.args[0]], a.length)
        if a.op == "BoolS":
            return claripy.BoolV(self.model[a.args[0]])
        if a.op == "FPS":
            return claripy.FPV(self.model[a.args[0]], a.args[1])
        if a.op == "StringS":
            return claripy.StringV(self.model[a.args[0]])
        return a

    def eval_ast(self, ast, allow_unconstrained: bool = True):
        """
        Eval the ast, replacing symbols by their last value in the model.

        :param ast:                 The AST to evaluate.
        :param allow_unconstrained: When set to True, we will treat non-existent variables as unconstrained variables
                                    and will use arbitrary concrete values for them during evaluation. Otherwise, raise
                                    KeyErrors for non-existent variables.
        """

        if allow_unconstrained:
            new_ast = claripy.replace_dict(ast, self.replacements, leaf_operation=self._leaf_op)
        else:
            new_ast = claripy.replace_dict(
                ast, self.constraint_only_replacements, leaf_operation=self._leaf_op_existonly
            )
        return backends.concrete.eval(new_ast, 1)[0]

    def eval_constraints(self, constraints):
        """Returns whether the constraints is satisfied trivially by using the
        last model."""
        # eval_ast is concretizing symbols and evaluating them, this can raise
        # exceptions.
        try:
            return all(self.eval_ast(c) for c in constraints)
        except errors.ClaripyZeroDivisionError:
            return False

    def eval_list(self, asts, allow_unconstrained: bool = True) -> tuple:
        """
        Evaluate a list of ASTs.

        :param asts:                A list of ASTs to evaluate.
        :param allow_unconstrained: When set to True, we will treat non-existent variables as unconstrained variables
                                    and will use arbitrary concrete values for them during evaluation. Otherwise, raise
                                    KeyErrors for non-existent variables.
        :return:                    A tuple of evaluated results, one element per AST.
        """

        return tuple(self.eval_ast(c, allow_unconstrained=allow_unconstrained) for c in asts)


class ModelCacheMixin:
    """ModelCacheMixin is a mixin that caches models for a solver."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._models = set()
        self._exhausted = False
        self._eval_exhausted = weakref.WeakValueDictionary()
        self._max_exhausted = weakref.WeakValueDictionary()
        self._min_exhausted = weakref.WeakValueDictionary()
        self._max_signed_exhausted = weakref.WeakValueDictionary()
        self._min_signed_exhausted = weakref.WeakValueDictionary()

    def _blank_copy(self, c):
        super()._blank_copy(c)
        c._models = set()
        c._exhausted = False
        c._eval_exhausted = weakref.WeakValueDictionary()
        c._max_exhausted = weakref.WeakValueDictionary()
        c._min_exhausted = weakref.WeakValueDictionary()
        c._max_signed_exhausted = weakref.WeakValueDictionary()
        c._min_signed_exhausted = weakref.WeakValueDictionary()

    def _copy(self, c):
        super()._copy(c)
        c._models = set(self._models)
        c._exhausted = self._exhausted
        c._eval_exhausted = weakref.WeakValueDictionary(self._eval_exhausted)
        c._max_exhausted = weakref.WeakValueDictionary(self._max_exhausted)
        c._min_exhausted = weakref.WeakValueDictionary(self._min_exhausted)
        c._max_signed_exhausted = weakref.WeakValueDictionary(self._max_signed_exhausted)
        c._min_signed_exhausted = weakref.WeakValueDictionary(self._min_signed_exhausted)

    def __setstate__(self, base_state):
        super().__setstate__(base_state)
        self._models = set()
        self._exhausted = False
        self._eval_exhausted = weakref.WeakValueDictionary()
        self._max_exhausted = weakref.WeakValueDictionary()
        self._min_exhausted = weakref.WeakValueDictionary()
        self._max_signed_exhausted = weakref.WeakValueDictionary()
        self._min_signed_exhausted = weakref.WeakValueDictionary()

    #
    # Model cleaning
    #

    def simplify(self):
        results = super().simplify()
        if len(results) > 0 and any(c is false() for c in results):
            self._models.clear()
        return results

    def _trivial_model_optimization(self):
        c = self.constraints[0]
        if not (
            c.depth == 2
            and c.op == "__eq__"
            and len(c.variables) == 1
            and c.args[0].symbolic
            and not c.args[1].symbolic
            and c.args[0].op == "BVS"
        ):
            return

        self._models.add(ModelCache({next(iter(c.args[0].variables)): backends.concrete.eval(c.args[1], 1)[0]}))
        self._eval_exhausted[c.args[0].hash()] = c.args[0]
        self._max_exhausted[c.args[0].hash()] = c.args[0]
        self._min_exhausted[c.args[0].hash()] = c.args[0]
        self._max_signed_exhausted[c.args[0].hash()] = c.args[0]
        self._min_signed_exhausted[c.args[0].hash()] = c.args[0]

    def _add(self, constraints, invalidate_cache=True):
        if len(constraints) == 0:
            return constraints

        old_vars = frozenset(self.variables)
        added = super()._add(constraints, invalidate_cache=invalidate_cache)
        if len(added) == 0:
            return added

        if len(self.constraints) == 1 and len(self._models) == 0:
            self._trivial_model_optimization()

        new_vars = any(a.variables - old_vars for a in added)
        if new_vars or invalidate_cache:
            # shortcut for unsat
            if any(c is false() for c in constraints):
                self._models.clear()

            still_valid = set(self._get_models(extra_constraints=added))
            if len(still_valid) != len(self._models):
                self._exhausted = False
                self._eval_exhausted.clear()
                self._max_exhausted.clear()
                self._min_exhausted.clear()
                self._max_signed_exhausted.clear()
                self._min_signed_exhausted.clear()
                self._models = still_valid

        return added

    def split(self):
        results = super().split()
        for r in results:
            r._models = {m.filter(r.variables) for m in self._models}
        return results

    def combine(self, others):
        combined = super().combine(others)

        if any(len(o._models) == 0 for o in others) or len(self._models) == 0:
            # this would need a solve anyways, so screw it
            return combined

        vars_count = len(self.variables) + sum(len(s.variables) for s in others)
        all_vars = self.variables.union(*[s.variables for s in others])
        if vars_count != len(all_vars):
            # this is the case where there are variables missing from the models.
            # We'll need more intelligence here to handle it
            return combined

        model_lists = [self._models]
        model_lists.extend(o._models for o in others)
        combined._models.update(
            itertools.starmap(ModelCache.combine, itertools.islice(itertools.product(*model_lists), len(self._models)))
        )
        return combined

    def update(self, other):
        """
        Updates this cache mixin with results discovered by the other split off one.
        """

        acceptable_models = [m for m in other._models if set(m.model.keys()) == self.variables]
        self._models.update(acceptable_models)
        self._eval_exhausted.update(other._eval_exhausted)
        self._max_exhausted.update(other._max_exhausted)
        self._min_exhausted.update(other._min_exhausted)
        self._max_signed_exhausted.update(other._max_signed_exhausted)
        self._min_signed_exhausted.update(other._min_signed_exhausted)

    #
    # Cache retrieval
    #

    def _model_hook(self, m):
        # Z3 might give us solutions for variables that we did not ask for. so we create a new dict with solutions for
        # only the variables that are under the solver's control
        m_ = {k: v for k, v in m.items() if k in self.variables}
        if m_:
            model = ModelCache(m_)
            self._models.add(model)

    def _get_models(self, extra_constraints=()):
        for m in self._models:
            if m.eval_constraints(extra_constraints):
                yield m

    def _get_batch_solutions(self, asts, n=None, extra_constraints=(), allow_unconstrained=True):
        results = set()

        for m in self._get_models(extra_constraints):
            try:
                results.add(m.eval_list(asts, allow_unconstrained=allow_unconstrained))
            except (ZeroDivisionError, KeyError):
                continue
            if len(results) == n:
                break

        return results

    def _get_solutions(self, e, n=None, extra_constraints=(), allow_unconstrained=True):
        return tuple(
            v[0]
            for v in self._get_batch_solutions(
                [e],
                n=n,
                extra_constraints=extra_constraints,
                allow_unconstrained=allow_unconstrained,
            )
        )

    #
    # Cached functions
    #

    def satisfiable(self, extra_constraints=(), exact=None):
        for _ in self._get_models(extra_constraints=extra_constraints):
            return True
        return super().satisfiable(extra_constraints=extra_constraints, exact=exact)

    def batch_eval(self, asts, n, extra_constraints=(), exact=None):
        results = self._get_batch_solutions(asts, n=n, extra_constraints=extra_constraints)

        if len(results) == n or (len(asts) == 1 and asts[0].hash() in self._eval_exhausted):
            return results

        remaining = n - len(results)

        # TODO: faster to concat?
        if len(results) != 0:
            constraints = (
                claripy.And(*[claripy.Or(*[a != v for a, v in zip(asts, r, strict=False)]) for r in results]),
                *tuple(extra_constraints),
            )
        else:
            constraints = extra_constraints

        try:
            results.update(super().batch_eval(asts, remaining, extra_constraints=constraints, exact=exact))
        except UnsatError:
            if len(results) == 0:
                raise

        if len(extra_constraints) == 0 and len(results) < n:
            for e in asts:
                # only mark an AST as eval-exhausted if e.variables is a subset of variables that the current solver
                # knows about (from its constraints)
                if self.variables.issuperset(e.variables):
                    self._eval_exhausted[e.hash()] = e

        return results

    def eval(self, e, n, extra_constraints=(), exact=None):
        return tuple(
            r[0] for r in ModelCacheMixin.batch_eval(self, [e], n, extra_constraints=extra_constraints, exact=exact)
        )

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        cached = []
        if e.hash() in self._eval_exhausted or e.hash() in self._min_exhausted:
            # we set allow_unconstrained to False because we expect all returned values for e are returned by Z3,
            # instead of some arbitrarily assigned concrete values.
            cached = self._get_solutions(e, extra_constraints=extra_constraints, allow_unconstrained=False)

        if len(cached) > 0:

            def signed_key(v):
                return v if v >= 0 else v + 2 ** len(e)

            return min(cached, key=signed_key if signed else lambda v: v)

        m = super().min(e, extra_constraints=extra_constraints, signed=signed, exact=exact)
        if len(extra_constraints) == 0:
            (self._min_signed_exhausted if signed else self._min_exhausted)[e.hash()] = e
        return m

    def max(self, e, extra_constraints=(), signed=False, exact=None):
        cached = []
        if e.hash() in self._eval_exhausted or e.hash() in self._max_exhausted:
            cached = self._get_solutions(e, extra_constraints=extra_constraints, allow_unconstrained=False)

        if len(cached) > 0:

            def signed_key(v):
                return v if v < 2 ** (len(e) - 1) else v - 2 ** len(e)

            return max(cached, key=signed_key if signed else lambda v: v)

        m = super().max(e, extra_constraints=extra_constraints, signed=signed, exact=exact)
        if len(extra_constraints) == 0:
            (self._max_signed_exhausted if signed else self._max_exhausted)[e.hash()] = e
        return m

    def solution(self, e, v, extra_constraints=(), exact=None):
        if isinstance(v, Base):
            cached = self._get_batch_solutions([e, v], extra_constraints=extra_constraints)
            if any(ec == vc for ec, vc in cached):
                return True
        else:
            cached = self._get_solutions(e, extra_constraints=extra_constraints)
            if v in cached:
                return True

        return super().solution(e, v, extra_constraints=extra_constraints, exact=exact)



from __future__ import annotations


class SolveBlockMixin:
    """SolveBlockMixin that adds the ability to block solving by setting can_solve to False."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.can_solve = True

    def _blank_copy(self, c):
        super()._blank_copy(c)
        c.can_solve = True

    def _copy(self, c):
        super()._copy(c)
        c.can_solve = self.can_solve

    def __getstate__(self):
        return self.can_solve, super().__getstate__()

    def __setstate__(self, s):
        self.can_solve, base_state = s
        super().__setstate__(base_state)

    def eval(self, e, n, extra_constraints=(), exact=None):
        assert self.can_solve
        return super().eval(e, n, extra_constraints=extra_constraints, exact=exact)

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None):
        assert self.can_solve
        return super().batch_eval(exprs, n, extra_constraints=extra_constraints, exact=exact)

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        assert self.can_solve
        return super().min(e, extra_constraints=extra_constraints, signed=signed, exact=exact)

    def max(self, e, extra_constraints=(), signed=False, exact=None):
        assert self.can_solve
        return super().max(e, extra_constraints=extra_constraints, signed=signed, exact=exact)

    def satisfiable(self, extra_constraints=(), exact=None):
        assert self.can_solve
        return super().satisfiable(extra_constraints=extra_constraints, exact=exact)

    def solution(self, e, v, extra_constraints=(), exact=None):
        assert self.can_solve
        return super().solution(e, v, extra_constraints=extra_constraints, exact=exact)



from __future__ import annotations

import ctypes
import logging
import numbers
import operator
import os
import signal
import sys
import threading
import weakref
from decimal import Decimal
from functools import reduce
from typing import TYPE_CHECKING, Any

import z3
from cachetools import LRUCache

import claripy
from claripy.ast import BV, FP, Bool
from claripy.backends.backend import Backend
from claripy.errors import (
    BackendError,
    ClaripyError,
    ClaripyOperationError,
    ClaripySolverInterruptError,
    ClaripyZ3Error,
)
from claripy.fp import RM, FSort
from claripy.operations import backend_fp_operations, backend_operations, backend_strings_operations, bound_ops

if TYPE_CHECKING:
    from types import FrameType

    from claripy.annotation import Annotation


log = logging.getLogger(__name__)

# pylint:disable=unidiomatic-typecheck

ALL_Z3_CONTEXTS = weakref.WeakSet()
INT_STRING_CHUNK_SIZE: int | None = None  # will be updated later if we are on CPython 3.11+


def handle_sigint(signals: int, frametype: FrameType | None) -> Any:
    if old_handler == signal.SIG_IGN:
        return

    contexts = list(ALL_Z3_CONTEXTS)
    for context in contexts:
        context.interrupt()

    if callable(old_handler):
        old_handler(signals, frametype)
    else:
        print("*** CRITICAL ERROR - THIS SHOULD NEVER HAPPEN", sys.stderr, flush=True)
        raise KeyboardInterrupt


if threading.current_thread() == threading.main_thread():  # Signal only works in the main thread
    old_handler = signal.getsignal(signal.SIGINT)
    if old_handler is None or old_handler == signal.SIG_DFL:
        # there is a signal handler installed by someone other than python. we cannot handle this.
        old_handler = True
    else:
        signal.signal(signal.SIGINT, handle_sigint)

try:
    import __pypy__

    _is_pypy = True
except ImportError:
    _is_pypy = False


def z3_expr_to_smt2(f, status="unknown", name="benchmark", logic=""):
    # from https://stackoverflow.com/a/14629021/9719920
    v = (z3.Ast * 0)()
    return z3.Z3_benchmark_to_smtlib_string(f.ctx_ref(), name, logic, status, "", 0, v, f.as_ast())


def claripy_solver_to_smt2(s):
    return s._get_solver().to_smt2()


def _add_memory_pressure(p):
    """
    PyPy's garbage collector is not aware of memory uses happening inside native code. When performing memory-intensive
    tasks in native code, the memory pressure that PyPy observes can greatly deviate from the actual memory pressure.
    We must manually add sufficient memory pressure to account for the "missing" portion.

    This is not a problem for CPython since its GC is based on reference counting.
    """

    global _is_pypy  # pylint:disable=global-variable-not-assigned
    if _is_pypy:
        __pypy__.add_memory_pressure(p)


def int_to_str_unlimited(v: int) -> str:
    """
    Convert an integer to a decimal string, without any size limit.

    :param v: The integer to convert.
    :return: The string.
    """

    if INT_STRING_CHUNK_SIZE is None:
        return str(v)

    if v == 0:
        return "0"

    MOD = 10**INT_STRING_CHUNK_SIZE
    v_str = ""
    if v < 0:
        is_negative = True
        v = -v
    else:
        is_negative = False
    while v > 0:
        v_chunk = str(v % MOD)
        v //= MOD
        if v > 0:
            v_chunk = v_chunk.zfill(INT_STRING_CHUNK_SIZE)
        v_str = v_chunk + v_str
    return v_str if not is_negative else "-" + v_str


def Z3_to_int_str(val):
    # we will monkey-patch Z3 and replace Z3._to_int_str with this version, which is free of integer size limits.

    if isinstance(val, float):
        return str(int(val))
    if isinstance(val, bool):
        return "1" if val else "0"
    return int_to_str_unlimited(val)


if hasattr(sys, "get_int_max_str_digits"):
    # CPython 3.11+
    # monkey-patch Z3 so that it can accept long integers
    z3.z3._to_int_str = Z3_to_int_str
    # update INT_STRING_CHUNK_SIZE
    INT_STRING_CHUNK_SIZE = sys.get_int_max_str_digits()


def str_to_int_unlimited(s: str) -> int:
    """
    Convert a decimal string to an integer, without any size limit.

    :param s: The string to convert.
    :return: The integer.
    """
    if INT_STRING_CHUNK_SIZE is None:
        return int(s)

    if not s:
        return int(s)  # an exception will be raised, which is intentional

    v = 0
    if s[0] == "-":
        is_negative = True
        s = s[1:]
    else:
        is_negative = False

    for i in range(0, len(s), INT_STRING_CHUNK_SIZE):
        start = i
        end = min(i + INT_STRING_CHUNK_SIZE, len(s))
        v *= 10 ** (end - start)
        v += int(s[start:end], 10)
    return v if not is_negative else -v


#
# Some global variables
#

# as of z3 4.8.7.0 this seems to matter
z3.set_param("rewriter.hi_fp_unspecified", "true")

#
# Utility functions
#


def condom(f):
    def z3_condom(*args, **kwargs):
        """
        The Z3 condom intercepts Z3Exceptions and throws a ClaripyZ3Error instead.
        """
        try:
            return f(*args, **kwargs)
        except z3.Z3Exception as ze:
            raise ClaripyZ3Error from ze

    return z3_condom


def _z3_decl_name_str(ctx, decl):
    decl_name = z3.Z3_get_decl_name(ctx, decl)
    return z3.Z3_get_symbol_string_bytes(ctx, decl_name)


def z3_solver_sat(solver, extra_constraints, occasion):
    log.debug("Doing a check! (%s)", occasion)

    result = solver.check(extra_constraints)

    if result == z3.unknown:
        reason = solver.reason_unknown()
        if reason in ("interrupted from keyboard", "interrupted"):
            raise KeyboardInterrupt
        if reason in ("timeout", "max. resource limit exceeded", "max. memory exceeded"):
            raise ClaripySolverInterruptError(reason)
        raise ClaripyZ3Error("solver unknown: " + reason)
    return result == z3.sat


class SmartLRUCache(LRUCache):
    def __init__(self, maxsize, getsizeof=None, evict=None):
        LRUCache.__init__(self, maxsize, getsizeof=getsizeof)
        self._evict = evict

    def popitem(self):
        key, val = LRUCache.popitem(self)
        if self._evict:
            self._evict(key, val)
        return key, val


#
# And the (ugh) magic
#


class BackendZ3(Backend):
    _split_on = ("And", "Or")

    def __init__(self, reuse_z3_solver=None, ast_cache_size=10000):
        Backend.__init__(self, solver_required=True)

        # Per-thread Z3 solver
        # This setting is treated as a global setting and is not supposed to be changed during runtime, unless you know
        # what you are doing.
        if reuse_z3_solver is None:
            reuse_z3_solver = os.environ.get("REUSE_Z3_SOLVER", "False").lower() in {"1", "true", "yes", "y"}
        self.reuse_z3_solver = reuse_z3_solver

        self._ast_cache_size = ast_cache_size

        # and the operations
        all_ops = backend_fp_operations | backend_operations
        all_ops |= backend_strings_operations - {"StrIsDigit"}
        for o in all_ops - {"BVV", "BoolV", "FPV", "FPS", "BitVec", "StringV"}:
            self._op_raw[o] = getattr(self, "_op_raw_" + o)
        self._op_raw["Xor"] = self._op_raw_Xor

        self._op_raw["UGE"] = self._op_raw_UGE
        self._op_raw["UGT"] = self._op_raw_UGT
        self._op_raw["ULE"] = self._op_raw_ULE
        self._op_raw["ULT"] = self._op_raw_ULT

        self._op_raw["Reverse"] = self._op_raw_Reverse
        self._op_raw["fpToSBV"] = self._op_raw_fpToSBV
        self._op_raw["fpToUBV"] = self._op_raw_fpToUBV

        self._op_expr["BVS"] = self.BVS
        self._op_expr["BVV"] = self.BVV
        self._op_expr["FPV"] = self.FPV
        self._op_expr["FPS"] = self.FPS
        self._op_expr["BoolV"] = self.BoolV
        self._op_expr["BoolS"] = self.BoolS
        self._op_expr["StringV"] = self.StringV
        self._op_expr["StringS"] = self.StringS

        self._op_raw["__floordiv__"] = self._op_div
        self._op_raw["__mod__"] = self._op_mod

        # reduceable
        self._op_raw["__add__"] = self._op_add
        self._op_raw["__sub__"] = self._op_sub
        self._op_raw["__mul__"] = self._op_mul
        self._op_raw["__or__"] = self._op_or
        self._op_raw["__xor__"] = self._op_xor
        self._op_raw["__and__"] = self._op_and

        self.solve_count = 0

    # This is necessary to be able to carry over annotations when round-tripping through Z3
    @property
    def bvs_annotations(self) -> dict[bytes, tuple[Annotation, ...]]:
        if not hasattr(self._tls, "bvs_annotations"):
            self._tls.bvs_annotations = {}
        return self._tls.bvs_annotations

    @property
    def _c_uint64_p(self):
        try:
            return self._tls.c_uint64_p
        except AttributeError:
            # a pointer to get values out of Z3
            self._tls.c_uint64_p = ctypes.pointer(ctypes.c_uint64())

            return self._tls.c_uint64_p

    @property
    def _context(self):
        try:
            return self._tls.context
        except AttributeError:
            main_thread = threading.current_thread() == threading.main_thread()
            self._tls.context = z3.Context() if not main_thread else z3.main_ctx()
            ALL_Z3_CONTEXTS.add(self._tls.context)
            return self._tls.context

    @property
    def _boolref_tactics(self):
        try:
            return self._tls.boolref_tactics
        except AttributeError:
            tactics = z3.Then(
                z3.Tactic("simplify", ctx=self._context),
                z3.Tactic("propagate-ineqs", ctx=self._context),
                z3.Tactic("propagate-values", ctx=self._context),
                z3.Tactic("unit-subsume-simplify", ctx=self._context),
                z3.Tactic("aig", ctx=self._context),
                ctx=self._context,
            )
            self._tls.boolref_tactics = tactics
            return self._tls.boolref_tactics

    @property
    def _ast_cache(self):
        try:
            return self._tls.ast_cache
        except AttributeError:
            self._tls.ast_cache = SmartLRUCache(self._ast_cache_size, evict=self._pop_from_ast_cache)
            return self._tls.ast_cache

    @property
    def _var_cache(self):
        try:
            return self._tls.var_cache
        except AttributeError:
            self._tls.var_cache = weakref.WeakValueDictionary()
            return self._tls.var_cache

    @property
    def _sym_cache(self):
        try:
            return self._tls.sym_cache
        except AttributeError:
            self._tls.sym_cache = weakref.WeakValueDictionary()
            return self._tls.sym_cache

    def downsize(self):
        Backend.downsize(self)

        self._ast_cache.clear()
        self._var_cache.clear()
        self._sym_cache.clear()

    def _name(self, o):  # pylint:disable=unused-argument
        log.warning("BackendZ3.name() called. This is weird.")
        raise BackendError("name is not implemented yet")

    def _pop_from_ast_cache(self, _, tpl):
        _, raw_ast = tpl
        z3.Z3_dec_ref(self._context.ctx, raw_ast)

    #
    # Core creation methods
    #

    @condom
    def BVS(self, ast):
        name = ast._encoded_name
        self.bvs_annotations[name] = ast.annotations
        size = ast.size()
        # TODO: Here we can use low level APIs because the check performed by the high level API always results in
        #       the else branch of the check. This evidence although comes from the execution of the angr and claripy
        #       test suite so I'm not sure if this assumption would hold on 100% of the cases
        bv = z3.BitVecSortRef(z3.Z3_mk_bv_sort(self._context.ref(), size), self._context)
        return z3.BitVecRef(
            z3.Z3_mk_const(self._context.ref(), z3.to_symbol(name, self._context), bv.ast), self._context
        )

    @condom
    def BVV(self, ast):
        if ast.args[0] is None:
            raise BackendError("Z3 can't handle empty BVVs")

        size = ast.size()
        # TODO: Here there is no need to use low level API since teh high level API just perform some conversions which
        #       are mandatory to fix the types of the arguments requested by the low level API
        return z3.BitVecVal(ast.args[0], size, ctx=self._context)

    @condom
    def FPS(self, ast):
        sort_z3 = self._convert(ast.args[1])
        # TODO: Here is safe to use low level API since the only condition the hig level API check is if the context is
        #       None and this never happens here
        return z3.FPRef(
            z3.Z3_mk_const(self._context.ref(), z3.to_symbol(ast._encoded_name, self._context), sort_z3.ast),
            self._context,
        )

    @condom
    def FPV(self, ast):
        val = str(ast.args[0])
        sort = self._convert(ast.args[1])
        if val in ("+oo", "+inf", "+Inf", "inf"):
            return z3.FPNumRef(z3.Z3_mk_fpa_inf(sort.ctx_ref(), sort.ast, False), sort.ctx)
        if val in ("-oo", "-inf", "-Inf"):
            return z3.FPNumRef(z3.Z3_mk_fpa_inf(sort.ctx_ref(), sort.ast, True), sort.ctx)
        if val in ("0.0", "+0.0"):
            return z3.FPNumRef(z3.Z3_mk_fpa_zero(sort.ctx_ref(), sort.ast, False), sort.ctx)
        if val == "-0.0":
            return z3.FPNumRef(z3.Z3_mk_fpa_zero(sort.ctx_ref(), sort.ast, True), sort.ctx)
        if val in ("NaN", "nan"):
            return z3.FPNumRef(z3.Z3_mk_fpa_nan(sort.ctx_ref(), sort.ast), sort.ctx)
        better_val = str(Decimal(ast.args[0]))
        return z3.FPNumRef(z3.Z3_mk_numeral(self._context.ref(), better_val, sort.ast), self._context)

    @condom
    def BoolS(self, ast):
        return z3.BoolRef(
            z3.Z3_mk_const(
                self._context.ref(), z3.to_symbol(ast._encoded_name, self._context), z3.BoolSort(self._context).ast
            ),
            self._context,
        )

    @condom
    def BoolV(self, ast):  # pylint:disable=unused-argument
        # TODO: Here the checks performed by the high level API are mandatory before calling the low level API
        #       So we can keep the high level API call here
        return z3.BoolVal(ast.args[0], ctx=self._context)

    @condom
    def StringV(self, ast):
        return z3.StringVal(ast.args[0], ctx=self._context)

    @condom
    def StringS(self, ast):
        return z3.String(ast.args[0], ctx=self._context)

    #
    # Conversions
    #

    @condom
    def _convert(self, obj):  # pylint:disable=arguments-renamed
        if isinstance(obj, FSort):
            return z3.FPSortRef(z3.Z3_mk_fpa_sort(self._context.ref(), obj.exp, obj.mantissa), self._context)
        if isinstance(obj, RM):
            if obj == RM.RM_NearestTiesEven:
                return z3.FPRMRef(z3.Z3_mk_fpa_round_nearest_ties_to_even(self._context.ref()), self._context)
            if obj == RM.RM_NearestTiesAwayFromZero:
                return z3.FPRMRef(z3.Z3_mk_fpa_round_nearest_ties_to_away(self._context.ref()), self._context)
            if obj == RM.RM_TowardsPositiveInf:
                return z3.FPRMRef(z3.Z3_mk_fpa_round_toward_positive(self._context.ref()), self._context)
            if obj == RM.RM_TowardsNegativeInf:
                return z3.FPRMRef(z3.Z3_mk_fpa_round_toward_negative(self._context.ref()), self._context)
            if obj == RM.RM_TowardsZero:
                return z3.FPRMRef(z3.Z3_mk_fpa_round_toward_zero(self._context.ref()), self._context)
            raise BackendError("unrecognized rounding mode")
        if obj is True:
            return z3.BoolRef(z3.Z3_mk_true(self._context.ref()), self._context)
        if obj is False:
            return z3.BoolRef(z3.Z3_mk_false(self._context.ref()), self._context)
        if isinstance(obj, numbers.Number | str) or (hasattr(obj, "__module__") and obj.__module__ in ("z3", "z3.z3")):
            return obj
        log.debug("BackendZ3 encountered unexpected type %s", type(obj))
        raise BackendError(f"unexpected type {type(obj)} encountered in BackendZ3")

    def call(self, *args, **kwargs):  # pylint;disable=arguments-renamed
        return Backend.call(self, *args, **kwargs)

    @condom
    def _abstract(self, e):
        return self._abstract_internal(e.ctx.ctx, e.ast)

    @staticmethod
    def _z3_ast_hash(ast):
        """
        This is a better hashing function for z3 Ast objects. Z3_get_ast_hash() creates too many hash collisions.

        :param ast: A z3 Ast object.
        :return:    An integer - the hash.
        """

        return ast.value

    def _abstract_internal(self, ctx, ast, split_on=None):
        h = self._z3_ast_hash(ast)
        try:
            cached_ast, _ = self._ast_cache[h]
            return cached_ast
        except KeyError:
            pass

        decl = z3.Z3_get_app_decl(ctx, ast)
        decl_num = z3.Z3_get_decl_kind(ctx, decl)
        z3_sort = z3.Z3_get_sort(ctx, ast)

        if decl_num not in z3_op_nums:
            raise ClaripyError(f"unknown decl kind {decl_num}")
        if op_map.get(z3_op_nums[decl_num]) is None:
            raise ClaripyError(f"unknown decl op {z3_op_nums[decl_num]}")
        op_name = op_map[z3_op_nums[decl_num]]

        num_args = z3.Z3_get_app_num_args(ctx, ast)
        split_on = self._split_on if split_on is None else split_on
        new_split_on = split_on if op_name in split_on else set()
        children = [self._abstract_internal(ctx, z3.Z3_get_app_arg(ctx, ast, i), new_split_on) for i in range(num_args)]

        append_children = True

        if op_name == "True":
            return claripy.true()
        if op_name == "False":
            return claripy.false()
        if op_name.startswith("RM_"):
            return RM(op_name)
        if op_name == "INTERNAL":
            return claripy.StringV(z3.SeqRef(ast).as_string())
        if op_name == "BitVecVal":
            bv_size = z3.Z3_get_bv_sort_size(ctx, z3_sort)
            if z3.Z3_get_numeral_uint64(ctx, ast, self._c_uint64_p):
                return claripy.BVV(self._c_uint64_p.contents.value, bv_size)
            bv_num = str_to_int_unlimited(z3.Z3_get_numeral_string(ctx, ast))
            return claripy.BVV(bv_num, bv_size)
        if op_name in ("FPVal", "MinusZero", "MinusInf", "PlusZero", "PlusInf", "NaN"):
            ebits = z3.Z3_fpa_get_ebits(ctx, z3_sort)
            sbits = z3.Z3_fpa_get_sbits(ctx, z3_sort)
            sort = FSort.from_params(ebits, sbits)
            val = self._abstract_fp_val(ctx, ast, op_name)
            return claripy.FPV(val, sort)
        if op_name == "fpToUBV":
            bv_size = z3.Z3_get_bv_sort_size(ctx, z3_sort)
            return claripy.fpToUBV(*children, bv_size)
        if op_name == "fpToSBV":
            bv_size = z3.Z3_get_bv_sort_size(ctx, z3_sort)
            return claripy.fpToSBV(*children, bv_size)
        if op_name == "SignExt":
            bv_size = z3.Z3_get_decl_int_parameter(ctx, decl, 0)
            return claripy.SignExt(bv_size, children[0])
        if op_name == "ZeroExt":
            bv_size = z3.Z3_get_decl_int_parameter(ctx, decl, 0)
            return claripy.ZeroExt(bv_size, children[0])
        if op_name == "RotateLeft":
            return claripy.RotateLeft(children[0], children[1])
        if op_name == "RotateRight":
            return claripy.RotateRight(children[0], children[1])

        if op_name == "UNINTERPRETED" and num_args == 0:  # symbolic value
            symbol_name = _z3_decl_name_str(ctx, decl)
            symbol_str = symbol_name.decode()
            symbol_ty = z3.Z3_get_sort_kind(ctx, z3_sort)

            if symbol_ty == z3.Z3_BV_SORT:
                bv_size = z3.Z3_get_bv_sort_size(ctx, z3_sort)
                annots = self.bvs_annotations.get(symbol_name, ())
                return claripy.BVS(symbol_str, bv_size, explicit_name=True, annotations=annots)
            if symbol_ty == z3.Z3_BOOL_SORT:
                return claripy.BoolS(symbol_str, explicit_name=True)
            if symbol_ty == z3.Z3_FLOATING_POINT_SORT:
                ebits = z3.Z3_fpa_get_ebits(ctx, z3_sort)
                sbits = z3.Z3_fpa_get_sbits(ctx, z3_sort)
                sort = FSort.from_params(ebits, sbits)
                return claripy.FPS(symbol_str, sort, explicit_name=True)
            if z3.Z3_is_string_sort(ctx, z3_sort):
                raise BackendError("Z3 backend does not support string symbols")
            raise BackendError(f"Unknown z3 term type {symbol_ty}...?")

        if op_name == "UNINTERPRETED":
            mystery_name = z3.Z3_get_symbol_string(ctx, z3.Z3_get_decl_name(ctx, decl))
            log.error("Mystery operation %s in BackendZ3._abstract_internal. Please report this.", mystery_name)
        elif op_name == "Extract":
            hi = z3.Z3_get_decl_int_parameter(ctx, decl, 0)
            lo = z3.Z3_get_decl_int_parameter(ctx, decl, 1)
            args = [hi, lo]
        elif op_name in ("SignExt", "ZeroExt"):
            num = z3.Z3_get_decl_int_parameter(ctx, decl, 0)
            args = [num]
        elif op_name in ("fpToFP", "fpToFPSigned", "fpToFPUnsigned"):
            exp = z3.Z3_fpa_get_ebits(ctx, z3_sort)
            mantissa = z3.Z3_fpa_get_sbits(ctx, z3_sort)
            sort = FSort.from_params(exp, mantissa)
            args = [*children, sort]
            append_children = False
        else:
            args = []

        if append_children:
            args.extend(children)

        # hmm.... honestly not sure what to do here
        result_ty = op_type_map[z3_op_nums[decl_num]]
        ty = type(args[-1])

        if isinstance(result_ty, str):
            err = (
                f"Unknown Z3 error in abstraction (result_ty == '{result_ty}'). "
                "Update your version of Z3, and, if the problem persists, open a claripy issue."
            )
            log.error(err)
            raise BackendError(err)

        if op_name == "If":
            # If is polymorphic and thus must be handled specially
            ty = type(args[1])

            a = ty("If", tuple(args), length=args[1].length)
        elif (
            op := getattr(ty, op_name, None)
            or getattr(result_ty, op_name, None)
            or (op_name in bound_ops and getattr(ty, bound_ops[op_name], None))
        ):
            if op.calc_length is not None:
                length = op.calc_length(*args)
                a = result_ty(op_name, tuple(args), length=length)
            else:
                a = result_ty(op_name, tuple(args))
        else:
            a = result_ty(op_name, tuple(args))

        self._ast_cache[h] = (a, ast)
        z3.Z3_inc_ref(ctx, ast)
        return a

    def _abstract_to_primitive(self, ctx, ast):
        decl = z3.Z3_get_app_decl(ctx, ast)
        decl_num = z3.Z3_get_decl_kind(ctx, decl)

        if decl_num not in z3_op_nums:
            raise ClaripyError(f"unknown decl kind {decl_num}")
        if op_map.get(z3_op_nums[decl_num]) is None:
            raise ClaripyError(f"unknown decl op {z3_op_nums[decl_num]}")
        op_name = op_map[z3_op_nums[decl_num]]

        if op_name == "BitVecVal":
            return self._abstract_bv_val(ctx, ast)
        if op_name == "True":
            return True
        if op_name == "False":
            return False
        if op_name in ("FPVal", "MinusZero", "MinusInf", "PlusZero", "PlusInf", "NaN"):
            return self._abstract_fp_val(ctx, ast, op_name)
        if op_name == "Concat":
            # Quirk in how z3 might handle NaN encodings - it will not give us a fully evaluated model
            # https://github.com/Z3Prover/z3/issues/518
            # this case will be triggered if the z3 rewriter.hi_fp_unspecified is set to true
            # update 6 jan 2020: idk if this is true anymore
            nargs = z3.Z3_get_app_num_args(ctx, ast)
            res = 0
            for i in range(nargs):
                arg_ast = z3.Z3_get_app_arg(ctx, ast, i)
                arg_decl = z3.Z3_get_app_decl(ctx, arg_ast)
                arg_decl_num = z3.Z3_get_decl_kind(ctx, arg_decl)
                arg_size = z3.Z3_get_bv_sort_size(ctx, z3.Z3_get_sort(ctx, arg_ast))

                neg = False
                if arg_decl_num == z3.Z3_OP_BNEG:
                    arg_ast = z3.Z3_get_app_arg(ctx, arg_ast, 0)
                    arg_decl = z3.Z3_get_app_decl(ctx, arg_ast)
                    arg_decl_num = z3.Z3_get_decl_kind(ctx, arg_decl)
                    neg = True
                if arg_decl_num != z3.Z3_OP_BNUM:
                    raise BackendError("Weird z3 model")

                arg_int = self._abstract_bv_val(ctx, arg_ast)
                if neg:
                    arg_int = (1 << arg_size) - arg_int
                res <<= arg_size
                res |= arg_int
            return res
        if op_name == "fpToIEEEBV":
            # Another quirk in the way z3 might handle nan encodings. see above
            # this case will be triggered if the z3 rewriter.hi_fp_unspecified is set to false
            # update 6 jan 2020: idk if this is true anymore
            arg_ast = z3.Z3_get_app_arg(ctx, ast, 0)
            return self._abstract_fp_encoded_val(ctx, arg_ast)
        if op_name == "INTERNAL":
            seq = z3.SeqRef(ast)
            if seq.is_string():
                return seq.as_string()
        raise BackendError("Unable to abstract Z3 object to primitive")

    def _abstract_bv_val(self, ctx, ast):
        if z3.Z3_get_numeral_uint64(ctx, ast, self._c_uint64_p):
            return self._c_uint64_p.contents.value
        return str_to_int_unlimited(z3.Z3_get_numeral_string(ctx, ast))

    @staticmethod
    def _abstract_fp_val(ctx, ast, op_name):
        if op_name == "FPVal":
            # TODO: do better than this
            fp_mantissa = float(z3.Z3_fpa_get_numeral_significand_string(ctx, ast))
            fp_exp = int(z3.Z3_fpa_get_numeral_exponent_string(ctx, ast, False))
            fp_sign_c = ctypes.c_int()  # pylint: disable=no-value-for-parameter
            z3.Z3_fpa_get_numeral_sign(ctx, ast, ctypes.byref(fp_sign_c))
            fp_sign = -1 if fp_sign_c.value != 0 else 1
            return fp_sign * fp_mantissa * (2**fp_exp)
        if op_name == "MinusZero":
            return -0.0
        if op_name == "MinusInf":
            return float("-inf")
        if op_name == "PlusZero":
            return 0.0
        if op_name == "PlusInf":
            return float("inf")
        if op_name == "NaN":
            return float("nan")
        raise BackendError("Called _abstract_fp_val with unknown type")

    @staticmethod
    def _abstract_fp_encoded_val(ctx, ast):
        decl = z3.Z3_get_app_decl(ctx, ast)
        decl_num = z3.Z3_get_decl_kind(ctx, decl)
        op_name = op_map[z3_op_nums[decl_num]]
        sort = z3.Z3_get_sort(ctx, ast)
        ebits = z3.Z3_fpa_get_ebits(ctx, sort)
        sbits = z3.Z3_fpa_get_sbits(ctx, sort) - 1  # includes sign bit

        if op_name == "FPVal":
            # TODO: do better than this
            fp_mantissa = int(z3.Z3_fpa_get_numeral_significand_string(ctx, ast))
            fp_exp = int(z3.Z3_fpa_get_numeral_exponent_string(ctx, ast, True))
            fp_sign_c = ctypes.c_int()  # pylint: disable=no-value-for-parameter
            z3.Z3_fpa_get_numeral_sign(ctx, ast, ctypes.byref(fp_sign_c))
            fp_sign = 1 if fp_sign_c.value != 0 else 0
        elif op_name == "MinusZero":
            fp_sign = 1
            fp_exp = 0
            fp_mantissa = 0
        elif op_name == "MinusInf":
            fp_sign = 1
            fp_exp = (1 << ebits) - 1
            fp_mantissa = 0
        elif op_name == "PlusZero":
            fp_sign = 0
            fp_exp = 0
            fp_mantissa = 0
        elif op_name == "PlusInf":
            fp_sign = 0
            fp_exp = (1 << ebits) - 1
            fp_mantissa = 0
        elif op_name == "NaN":
            fp_sign = 0
            fp_exp = (1 << ebits) - 1
            fp_mantissa = 1
        else:
            raise BackendError("Called _abstract_fp_val with unknown type")

        return (fp_sign << (ebits + sbits)) | (fp_exp << sbits) | fp_mantissa

    def solver(self, timeout=None, max_memory=None):
        if not self.reuse_z3_solver or getattr(self._tls, "solver", None) is None:
            s = z3.Solver(ctx=self._context)  # , logFile="claripy.smt2")
            if threading.current_thread() != threading.main_thread():
                s.set(ctrl_c=False)
            _add_memory_pressure(1024 * 1024 * 10)
            if self.reuse_z3_solver:
                # Store the Z3 solver to a thread-local storage if the reuse-solver option is enabled
                self._tls.solver = s
        else:
            # Load the existing Z3 solver for this thread
            s = self._tls.solver
            s.reset()

        # Configure timeouts
        if timeout is not None:
            if "soft_timeout" in str(s.param_descrs()):
                s.set("soft_timeout", timeout)
                s.set("solver2_timeout", timeout)
            else:
                s.set("timeout", timeout)
        if max_memory is not None:
            s.set("max_memory", max_memory)
        return s

    def clone_solver(self, s):
        # This clones the solver.
        # See https://github.com/Z3Prover/z3/issues/556
        return s.translate(self._context)

    def _add(self, s, c, track=False):
        if track:
            already_tracked = {str(impl.children()[0]) for impl in s.assertions()}
            for constraint in c:
                name = str(hash(constraint))
                if name not in already_tracked:
                    s.assert_and_track(constraint, name)
                    already_tracked.add(name)
        else:
            s.add(*c)

    def add(self, s, c, track=False):
        converted = self.convert_list(c)
        if track:
            for a, nice_ast in zip(c, converted, strict=False):
                ast = nice_ast.ast
                h = self._z3_ast_hash(ast)
                self._ast_cache[h] = (a, ast)
        return self._add(s, converted, track=track)

    def _unsat_core(self, s):
        cores = s.unsat_core()
        return [impl.children()[1] for impl in s.assertions() if impl.children()[0] in cores]

    @condom
    def _primitive_from_model(self, model, expr):
        v = model.eval(expr, model_completion=True)
        return self._abstract_to_primitive(v.ctx.ctx, v.ast)

    #
    # New, model-driven solves
    #

    def _generic_model(self, z3_model):
        """
        Converts a Z3 model to a name->primitive dict.
        """
        model = {}
        for m_f in z3_model:
            n = _z3_decl_name_str(m_f.ctx.ctx, m_f.ast).decode()
            m = m_f()
            me = z3_model.eval(m, model_completion=True)
            model[n] = self._abstract_to_primitive(me.ctx.ctx, me.ast)

        return model

    def _satisfiable(self, extra_constraints=(), solver=None, model_callback=None):
        self.solve_count += 1

        log.debug("Doing a check! (satisfiable)")
        if not z3_solver_sat(solver, extra_constraints, "satisfiable"):
            return False

        if model_callback is not None:
            model_callback(self._generic_model(solver.model()))
        return True

    def _eval(self, expr, n, extra_constraints=(), solver=None, model_callback=None):
        results = self._batch_eval(
            [expr], n, extra_constraints=extra_constraints, solver=solver, model_callback=model_callback
        )

        # Unpack it
        return [v[0] for v in results]

    @condom
    def _batch_eval(self, exprs, n, extra_constraints=(), solver=None, model_callback=None):
        result_values = []

        if n > 1:
            solver.push()

        for i in range(n):
            self.solve_count += 1
            if not z3_solver_sat(solver, extra_constraints, "batch_eval"):
                break
            model = solver.model()

            # construct results
            r = []
            for expr in exprs:
                if not isinstance(expr, numbers.Number | str | bool):
                    v = self._primitive_from_model(model, expr)
                    r.append(v)
                else:
                    r.append(expr)

            # Append the solution to the result list
            if model_callback is not None:
                model_callback(self._generic_model(solver.model()))
            result_values.append(tuple(r))

            # Construct the extra constraint so we don't get the same result anymore
            if i + 1 != n:
                if len(exprs) == 1:
                    solver.add(exprs[0] != r[0])
                else:
                    solver.add(
                        self._op_raw_Not(self._op_raw_And(*[(ex == ex_v) for ex, ex_v in zip(exprs, r, strict=False)]))
                    )
                model = None

        if n > 1:
            solver.pop()

        return result_values

    def _extrema(self, is_max: bool, expr, extra_constraints, signed, solver, model_callback):
        """
        _max if is_max else _min
        """
        lo = -(2 ** (expr.size() - 1)) if signed else 0
        hi = 2 ** (expr.size() - 1) - 1 if signed else 2 ** expr.size() - 1

        constraints = [self.convert(e) for e in extra_constraints]
        comment = "max" if is_max else "min"

        GE = operator.ge if signed else z3.UGE
        LE = operator.le if signed else z3.ULE

        # TODO: Can only deal with bitvectors, not floats
        while hi - lo > 1:
            middle = (lo + hi) // 2
            # l.debug("h/m/l/d: %d %d %d %d", hi, middle, lo, hi-lo)

            # TODO: is this assumption correct?
            # here it's not safe to call directly the z3 low level API since it might happen that the argument is an
            # integer and not a BV
            constraints.append(
                z3.And(GE(expr, middle), LE(expr, hi)) if is_max else z3.And(GE(expr, lo), LE(expr, middle))
            )

            self.solve_count += 1
            sat = z3_solver_sat(solver, constraints, comment)
            constraints.pop()
            if sat:
                log.debug("... still sat")
                if model_callback is not None:
                    model_callback(self._generic_model(solver.model()))
            else:
                log.debug("... now unsat")
            if sat == is_max:
                lo = middle
            else:
                hi = middle

        constraints.append(expr == (hi if is_max else lo))
        sat = z3_solver_sat(solver, constraints, comment)
        if sat and model_callback is not None:
            model_callback(self._generic_model(solver.model()))
        return hi if sat == is_max else lo

    @condom
    def _min(self, expr, extra_constraints=(), signed=False, solver=None, model_callback=None):
        return self._extrema(False, expr, extra_constraints, signed, solver, model_callback)

    @condom
    def _max(self, expr, extra_constraints=(), signed=False, solver=None, model_callback=None):
        return self._extrema(True, expr, extra_constraints, signed, solver, model_callback)

    @condom
    def simplify(self, expr):
        expr_raw = self.convert(expr)

        if isinstance(expr_raw, z3.BoolRef):
            boolref_tactics = self._boolref_tactics
            simplified = boolref_tactics(expr_raw).as_expr()
        else:
            simplified = z3.simplify(expr_raw)

        return self._abstract(simplified)

    def _is_false(self, e, extra_constraints=(), solver=None, model_callback=None):
        return z3.simplify(e).eq(z3.BoolVal(False, ctx=self._context))

    def _is_true(self, e, extra_constraints=(), solver=None, model_callback=None):
        return z3.simplify(e).eq(z3.BoolVal(True, ctx=self._context))

    def _solution(self, expr, v, extra_constraints=(), solver=None, model_callback=None):
        return self._satisfiable(
            extra_constraints=(expr == v, *tuple(extra_constraints)), solver=solver, model_callback=model_callback
        )

    #
    # Some Z3 passthroughs
    #

    # these require the context or special treatment

    @staticmethod
    def _op_div(a, b):
        return z3.BitVecRef(z3.Z3_mk_bvudiv(a.ctx_ref(), a.as_ast(), b.as_ast()), a.ctx)

    @staticmethod
    def _op_mod(a, b):
        return z3.BitVecRef(z3.Z3_mk_bvurem(a.ctx_ref(), a.as_ast(), b.as_ast()), a.ctx)

    @staticmethod
    def _op_add(*args):
        return reduce(operator.__add__, args)

    @staticmethod
    def _op_sub(*args):
        return reduce(operator.__sub__, args)

    @staticmethod
    def _op_mul(*args):
        return reduce(operator.__mul__, args)

    @staticmethod
    def _op_or(*args):
        return reduce(operator.__or__, args)

    @staticmethod
    def _op_xor(*args):
        return reduce(operator.__xor__, args)

    @staticmethod
    def _op_and(*args):
        return reduce(operator.__and__, args)

    def _op_raw_And(self, *args):
        # copied from z3._to_ast_array
        sz = len(args)
        _args = (z3.Ast * sz)()
        for i in range(sz):
            _args[i] = args[i].as_ast()
        return z3.BoolRef(z3.Z3_mk_and(self._context.ref(), sz, _args), self._context)

    def _op_raw_Xor(self, *args):
        return z3.BoolRef(z3.Z3_mk_xor(self._context.ref(), *(arg.as_ast() for arg in args)), self._context)

    def _op_raw_Or(self, *args):
        # copied from z3._to_ast_array
        sz = len(args)
        _args = (z3.Ast * sz)()
        for i in range(sz):
            _args[i] = args[i].as_ast()
        return z3.BoolRef(z3.Z3_mk_or(self._context.ref(), sz, _args), self._context)

    def _op_raw_Not(self, a):
        return z3.BoolRef(z3.Z3_mk_not(self._context.ref(), a.as_ast()), self._context)

    def _op_raw_If(self, i, t, e):
        # partially copied from z3._to_expr_ref
        ctx_ref = self._context.ref()
        ast = z3.Z3_mk_ite(ctx_ref, i.as_ast(), t.as_ast(), e.as_ast())
        k = z3.Z3_get_ast_kind(ctx_ref, ast)
        sk = z3.Z3_get_sort_kind(ctx_ref, z3.Z3_get_sort(ctx_ref, ast))
        if sk == z3.Z3_BOOL_SORT:
            return z3.BoolRef(ast, self._context)
        if sk == z3.Z3_BV_SORT:
            if k == z3.Z3_NUMERAL_AST:
                return z3.BitVecNumRef(ast, self._context)
            return z3.BitVecRef(ast, self._context)
        if sk == z3.Z3_FLOATING_POINT_SORT:
            if k == z3.Z3_APP_AST and z3.Z3_is_numeral_ast(ctx_ref, ast):
                return z3.FPNumRef(ast, self._context)
            return z3.FPRef(ast, self._context)
        return z3.ExprRef(ast, self._context)

    @condom
    def _op_raw_fpAbs(self, a):
        return z3.FPRef(z3.Z3_mk_fpa_abs(self._context.ref(), a.as_ast()), self._context)

    @condom
    def _op_raw_fpNeg(self, a):
        return z3.FPRef(z3.Z3_mk_fpa_neg(self._context.ref(), a.as_ast()), self._context)

    @condom
    def _op_raw_fpAdd(self, rm, a, b):
        return z3.FPRef(z3.Z3_mk_fpa_add(self._context.ref(), rm.as_ast(), a.as_ast(), b.as_ast()), self._context)

    @condom
    def _op_raw_fpSub(self, rm, a, b):
        return z3.FPRef(z3.Z3_mk_fpa_sub(self._context.ref(), rm.as_ast(), a.as_ast(), b.as_ast()), self._context)

    @condom
    def _op_raw_fpMul(self, rm, a, b):
        return z3.FPRef(z3.Z3_mk_fpa_mul(self._context.ref(), rm.as_ast(), a.as_ast(), b.as_ast()), self._context)

    @condom
    def _op_raw_fpDiv(self, rm, a, b):
        return z3.FPRef(z3.Z3_mk_fpa_div(self._context.ref(), rm.as_ast(), a.as_ast(), b.as_ast()), self._context)

    @condom
    def _op_raw_fpSqrt(self, rm, a):
        return z3.FPRef(z3.Z3_mk_fpa_sqrt(self._context.ref(), rm.as_ast(), a.as_ast()), self._context)

    @condom
    def _op_raw_fpLT(self, a, b):
        return z3.BoolRef(z3.Z3_mk_fpa_lt(self._context.ref(), a.as_ast(), b.as_ast()), self._context)

    @condom
    def _op_raw_fpLEQ(self, a, b):
        return z3.BoolRef(z3.Z3_mk_fpa_leq(self._context.ref(), a.as_ast(), b.as_ast()), self._context)

    @condom
    def _op_raw_fpGT(self, a, b):
        return z3.BoolRef(z3.Z3_mk_fpa_gt(self._context.ref(), a.as_ast(), b.as_ast()), self._context)

    @condom
    def _op_raw_fpGEQ(self, a, b):
        return z3.BoolRef(z3.Z3_mk_fpa_geq(self._context.ref(), a.as_ast(), b.as_ast()), self._context)

    @condom
    def _op_raw_fpEQ(self, a, b):
        return z3.BoolRef(z3.Z3_mk_fpa_eq(self._context.ref(), a.as_ast(), b.as_ast()), self._context)

    @condom
    def _op_raw_fpNEQ(self, a, b):
        return z3.Not(z3.BoolRef(z3.Z3_mk_fpa_eq(self._context.ref(), a.as_ast(), b.as_ast()), self._context))

    @condom
    def _op_raw_fpIsNaN(self, a):
        return z3.BoolRef(z3.Z3_mk_fpa_is_nan(self._context.ref(), a.as_ast()), self._context)

    @condom
    def _op_raw_fpIsInf(self, a):
        return z3.BoolRef(z3.Z3_mk_fpa_is_infinite(self._context.ref(), a.as_ast()), self._context)

    @condom
    def _op_raw_fpFP(self, sgn, exp, sig):
        return z3.FPRef(z3.Z3_mk_fpa_fp(self._context.ref(), sgn.ast, exp.ast, sig.ast), self._context)

    @condom
    def _op_raw_fpToSBV(self, rm, fp, bv_len):
        return z3.BitVecRef(z3.Z3_mk_fpa_to_sbv(self._context.ref(), rm.ast, fp.ast, bv_len), self._context)

    @condom
    def _op_raw_fpToUBV(self, rm, fp, bv_len):
        return z3.BitVecRef(z3.Z3_mk_fpa_to_ubv(self._context.ref(), rm.ast, fp.ast, bv_len), self._context)

    @condom
    def _op_raw_fpToFP(self, a1, a2=None, a3=None):
        # TODO: lots of mandatory checks are performed by the high level API here. we shouldn't use low level APIs here
        return z3.fpToFP(a1, a2=a2, a3=a3, ctx=self._context)

    @condom
    def _op_raw_fpToFPUnsigned(self, a1, a2, a3):
        # as above
        return z3.fpToFPUnsigned(a1, a2, a3, ctx=self._context)

    @condom
    def _op_raw_fpToIEEEBV(self, x):
        return z3.BitVecRef(z3.Z3_mk_fpa_to_ieee_bv(self._context.ref(), x.ast), self._context)

    # and these do not
    @staticmethod
    @condom
    def _op_raw_Concat(*args):
        sz = len(args)
        ctx = args[0].ctx
        # TODO: I don't think this is needed for us, we don't deal with Seq or Regex
        # if z3.is_seq(args[0]) or isinstance(args[0], str):
        #     v = (z3.Ast * sz)()
        #     for i in range(sz):
        #         v[i] = args[i].as_ast()
        #     return z3.SeqRef(z3.Z3_mk_seq_concat(ctx.ref(), sz, v), ctx)
        #
        # if z3.is_re(args[0]):
        #     v = (z3.Ast * sz)()
        #     for i in range(sz):
        #         v[i] = args[i].as_ast()
        #     return z3.ReRef(z3.Z3_mk_re_concat(ctx.ref(), sz, v), ctx)

        r = args[0]
        for i in range(sz - 1):
            r = z3.BitVecRef(z3.Z3_mk_concat(ctx.ref(), r.as_ast(), args[i + 1].as_ast()), ctx)
        return r

    @staticmethod
    @condom
    def _op_raw_Extract(high, low, a):
        # TODO: I don't think this is needed for us, we don't deal with Seq or Regex
        # if isinstance(high, str):
        #     high = z3.StringVal(high)
        # if z3.is_seq(high):
        #     s = high
        #     offset, length = _coerce_exprs(low, a, s.ctx)
        #     return z3.SeqRef(z3.Z3_mk_seq_extract(s.ctx_ref(), s.as_ast(), offset.as_ast(), length.as_ast()), s.ctx)
        return z3.BitVecRef(z3.Z3_mk_extract(a.ctx_ref(), high, low, a.as_ast()), a.ctx)

    @staticmethod
    @condom
    def _op_raw_LShR(a, b):
        return z3.BitVecRef(z3.Z3_mk_bvlshr(a.ctx_ref(), a.as_ast(), b.as_ast()), a.ctx)

    @staticmethod
    @condom
    def _op_raw_RotateLeft(a, b):
        return z3.BitVecRef(z3.Z3_mk_ext_rotate_left(a.ctx_ref(), a.as_ast(), b.as_ast()), a.ctx)

    @staticmethod
    @condom
    def _op_raw_RotateRight(a, b):
        return z3.BitVecRef(z3.Z3_mk_ext_rotate_right(a.ctx_ref(), a.as_ast(), b.as_ast()), a.ctx)

    @staticmethod
    @condom
    def _op_raw_SignExt(n, a):
        return z3.BitVecRef(z3.Z3_mk_sign_ext(a.ctx_ref(), n, a.as_ast()), a.ctx)

    @staticmethod
    @condom
    def _op_raw_UGE(a, b):
        return z3.BoolRef(z3.Z3_mk_bvuge(a.ctx_ref(), a.as_ast(), b.as_ast()), a.ctx)

    @staticmethod
    @condom
    def _op_raw_UGT(a, b):
        return z3.BoolRef(z3.Z3_mk_bvugt(a.ctx_ref(), a.as_ast(), b.as_ast()), a.ctx)

    @staticmethod
    @condom
    def _op_raw_ULE(a, b):
        return z3.BoolRef(z3.Z3_mk_bvule(a.ctx_ref(), a.as_ast(), b.as_ast()), a.ctx)

    @staticmethod
    @condom
    def _op_raw_ULT(a, b):
        return z3.BoolRef(z3.Z3_mk_bvult(a.ctx_ref(), a.as_ast(), b.as_ast()), a.ctx)

    @staticmethod
    @condom
    def _op_raw_ZeroExt(n, a):
        return z3.BitVecRef(z3.Z3_mk_zero_ext(a.ctx_ref(), n, a.as_ast()), a.ctx)

    @staticmethod
    @condom
    def _op_raw_SMod(a, b):
        return z3.BitVecRef(z3.Z3_mk_bvsrem(a.ctx_ref(), a.as_ast(), b.as_ast()), a.ctx)

    @staticmethod
    @condom
    def _op_raw_Reverse(a):
        if a.size() == 8:
            return a
        if a.size() % 8 != 0:
            raise ClaripyOperationError("can't reverse non-byte sized bitvectors")
        return BackendZ3._op_raw_Concat(*[BackendZ3._op_raw_Extract(i + 7, i, a) for i in range(0, a.size(), 8)])

    @staticmethod
    @condom
    def _op_raw_SLT(a, b):
        return a < b

    @staticmethod
    @condom
    def _op_raw_SLE(a, b):
        return a <= b

    @staticmethod
    @condom
    def _op_raw_SGT(a, b):
        return a > b

    @staticmethod
    @condom
    def _op_raw_SGE(a, b):
        return a >= b

    @staticmethod
    @condom
    def _op_raw_SDiv(a, b):
        return a / b

    def _identical(self, a, b):
        return a.eq(b)

    # String operations:

    @staticmethod
    @condom
    def _op_raw_StrConcat(*args):
        return z3.Concat(*args)

    @staticmethod
    @condom
    def _op_raw_StrSubstr(start_idx, count, initial_string):
        return z3.SubString(initial_string, z3.BV2Int(start_idx), z3.BV2Int(count))

    @staticmethod
    @condom
    def _op_raw_StrLen(input_string):
        return z3.Int2BV(z3.Length(input_string), 64)

    @staticmethod
    @condom
    def _op_raw_StrReplace(input_string, target, replacement):
        return z3.Replace(input_string, target, replacement)

    @staticmethod
    @condom
    def _op_raw_StrContains(input_string, substring):
        return z3.Contains(input_string, substring)

    @staticmethod
    @condom
    def _op_raw_StrPrefixOf(prefix, input_string):
        return z3.PrefixOf(prefix, input_string)

    @staticmethod
    @condom
    def _op_raw_StrSuffixOf(suffix, input_string):
        return z3.SuffixOf(suffix, input_string)

    @staticmethod
    @condom
    def _op_raw_StrIndexOf(string, pattern, start_idx):
        return z3.Int2BV(z3.IndexOf(string, pattern, z3.BV2Int(start_idx)), 64)

    @staticmethod
    @condom
    def _op_raw_StrToInt(input_string):
        return z3.Int2BV(z3.StrToInt(input_string), 64)

    @staticmethod
    @condom
    def _op_raw_IntToStr(input_bv):
        return z3.IntToStr(z3.BV2Int(input_bv))


#
# this is for the actual->abstract conversion above
#

# these are Z3 operation names for abstraction
z3_op_names = [_ for _ in dir(z3) if _.startswith("Z3_OP")]
z3_op_nums = {getattr(z3, _): _ for _ in z3_op_names}

op_map = {
    "Z3_OP_ADD": "__add__",
    "Z3_OP_AGNUM": None,
    "Z3_OP_AND": "And",
    "Z3_OP_ANUM": None,
    "Z3_OP_ARRAY_DEFAULT": None,
    "Z3_OP_ARRAY_EXT": None,
    "Z3_OP_ARRAY_MAP": None,
    "Z3_OP_AS_ARRAY": None,
    "Z3_OP_BADD": "__add__",
    "Z3_OP_BAND": "__and__",
    "Z3_OP_BASHR": "__rshift__",
    "Z3_OP_BCOMP": None,
    "Z3_OP_BIT0": None,
    "Z3_OP_BIT1": None,
    "Z3_OP_BIT2BOOL": None,
    "Z3_OP_BLSHR": "LShR",
    "Z3_OP_BMUL": "__mul__",
    "Z3_OP_BNAND": None,
    "Z3_OP_BNEG": "__neg__",
    "Z3_OP_BNOR": None,
    "Z3_OP_BNOT": "__invert__",
    "Z3_OP_BNUM": "BitVecVal",
    "Z3_OP_BOR": "__or__",
    "Z3_OP_BREDAND": None,
    "Z3_OP_BREDOR": None,
    "Z3_OP_BSDIV": "SDiv",
    "Z3_OP_BSDIV0": None,
    "Z3_OP_BSDIV_I": "SDiv",
    "Z3_OP_BSHL": "__lshift__",
    "Z3_OP_BSMOD": "SMod",
    "Z3_OP_BSMOD0": None,
    "Z3_OP_BSMOD_I": "SMod",
    "Z3_OP_BSMUL_NO_OVFL": None,
    "Z3_OP_BSMUL_NO_UDFL": None,
    "Z3_OP_BSREM": "SMod",
    "Z3_OP_BSREM0": None,
    "Z3_OP_BSREM_I": "SMod",
    "Z3_OP_BSUB": "__sub__",
    "Z3_OP_BUDIV": "__floordiv__",
    "Z3_OP_BUDIV0": None,
    "Z3_OP_BUDIV_I": "__floordiv__",
    "Z3_OP_BUMUL_NO_OVFL": None,
    "Z3_OP_BUREM": "__mod__",
    "Z3_OP_BUREM0": None,
    "Z3_OP_BUREM_I": "__mod__",
    "Z3_OP_BV2INT": None,
    "Z3_OP_BXNOR": None,
    "Z3_OP_BXOR": "__xor__",
    "Z3_OP_CARRY": None,
    "Z3_OP_CHAR_CONST": None,
    "Z3_OP_CHAR_FROM_BV": None,
    "Z3_OP_CHAR_IS_DIGIT": None,
    "Z3_OP_CHAR_LE": None,
    "Z3_OP_CHAR_TO_BV": None,
    "Z3_OP_CHAR_TO_INT": None,
    "Z3_OP_CONCAT": "Concat",
    "Z3_OP_CONST_ARRAY": None,
    "Z3_OP_DISTINCT": "__ne__",
    "Z3_OP_DIV": "SDiv",
    "Z3_OP_DT_ACCESSOR": None,
    "Z3_OP_DT_CONSTRUCTOR": None,
    "Z3_OP_DT_IS": None,
    "Z3_OP_DT_RECOGNISER": None,
    "Z3_OP_DT_UPDATE_FIELD": None,
    "Z3_OP_EQ": "__eq__",
    "Z3_OP_EXTRACT": "Extract",
    "Z3_OP_EXT_ROTATE_LEFT": "RotateLeft",
    "Z3_OP_EXT_ROTATE_RIGHT": "RotateRight",
    "Z3_OP_FALSE": "False",
    "Z3_OP_FD_CONSTANT": None,
    "Z3_OP_FD_LT": None,
    "Z3_OP_FPA_ABS": "fpAbs",
    "Z3_OP_FPA_ADD": "fpAdd",
    "Z3_OP_FPA_BV2RM": None,
    "Z3_OP_FPA_BVWRAP": None,
    "Z3_OP_FPA_DIV": "fpDiv",
    "Z3_OP_FPA_EQ": "fpEQ",
    "Z3_OP_FPA_FMA": None,
    "Z3_OP_FPA_FP": None,
    "Z3_OP_FPA_GE": "fpGEQ",
    "Z3_OP_FPA_GT": "fpGT",
    "Z3_OP_FPA_IS_INF": None,
    "Z3_OP_FPA_IS_NAN": None,
    "Z3_OP_FPA_IS_NEGATIVE": None,
    "Z3_OP_FPA_IS_NORMAL": None,
    "Z3_OP_FPA_IS_POSITIVE": None,
    "Z3_OP_FPA_IS_SUBNORMAL": None,
    "Z3_OP_FPA_IS_ZERO": None,
    "Z3_OP_FPA_LE": "fpLEQ",
    "Z3_OP_FPA_LT": "fpLT",
    "Z3_OP_FPA_MAX": None,
    "Z3_OP_FPA_MIN": None,
    "Z3_OP_FPA_MINUS_INF": "MinusInf",
    "Z3_OP_FPA_MINUS_ZERO": "MinusZero",
    "Z3_OP_FPA_MUL": "fpMul",
    "Z3_OP_FPA_NAN": "NaN",
    "Z3_OP_FPA_NEG": "fpNeg",
    "Z3_OP_FPA_NUM": "FPVal",
    "Z3_OP_FPA_PLUS_INF": "PlusInf",
    "Z3_OP_FPA_PLUS_ZERO": "PlusZero",
    "Z3_OP_FPA_REM": None,
    "Z3_OP_FPA_RM_NEAREST_TIES_TO_AWAY": "RM_RNA",
    "Z3_OP_FPA_RM_NEAREST_TIES_TO_EVEN": "RM_RNE",
    "Z3_OP_FPA_RM_TOWARD_NEGATIVE": "RM_RTN",
    "Z3_OP_FPA_RM_TOWARD_POSITIVE": "RM_RTP",
    "Z3_OP_FPA_RM_TOWARD_ZERO": "RM_RTZ",
    "Z3_OP_FPA_ROUND_TO_INTEGRAL": None,
    "Z3_OP_FPA_SQRT": "fpSqrt",
    "Z3_OP_FPA_SUB": "fpSub",
    "Z3_OP_FPA_TO_FP": "fpToFP",
    "Z3_OP_FPA_TO_FP_UNSIGNED": "fpToFPUnsigned",
    "Z3_OP_FPA_TO_IEEE_BV": "fpToIEEEBV",
    "Z3_OP_FPA_TO_REAL": None,
    "Z3_OP_FPA_TO_SBV": "fpToSBV",
    "Z3_OP_FPA_TO_UBV": "fpToUBV",
    "Z3_OP_GE": "SGE",
    "Z3_OP_GT": "SGT",
    "Z3_OP_IDIV": "SDiv",
    "Z3_OP_IFF": "__eq__",
    "Z3_OP_IMPLIES": None,
    "Z3_OP_INT2BV": None,
    "Z3_OP_INTERNAL": "INTERNAL",
    "Z3_OP_INT_TO_STR": None,
    "Z3_OP_IS_INT": None,
    "Z3_OP_ITE": "If",
    "Z3_OP_LABEL": None,
    "Z3_OP_LABEL_LIT": None,
    "Z3_OP_LE": "SLE",
    "Z3_OP_LT": "SLT",
    "Z3_OP_MOD": "__mod__",
    "Z3_OP_MUL": "__mul__",
    "Z3_OP_NOT": "Not",
    "Z3_OP_OEQ": None,
    "Z3_OP_OR": "Or",
    "Z3_OP_PB_AT_LEAST": None,
    "Z3_OP_PB_AT_MOST": None,
    "Z3_OP_PB_EQ": None,
    "Z3_OP_PB_GE": None,
    "Z3_OP_PB_LE": None,
    "Z3_OP_POWER": None,
    "Z3_OP_PR_AND_ELIM": None,
    "Z3_OP_PR_APPLY_DEF": None,
    "Z3_OP_PR_ASSERTED": None,
    "Z3_OP_PR_ASSUMPTION_ADD": None,
    "Z3_OP_PR_BIND": None,
    "Z3_OP_PR_CLAUSE_TRAIL": None,
    "Z3_OP_PR_COMMUTATIVITY": None,
    "Z3_OP_PR_DEF_AXIOM": None,
    "Z3_OP_PR_DEF_INTRO": None,
    "Z3_OP_PR_DER": None,
    "Z3_OP_PR_DISTRIBUTIVITY": None,
    "Z3_OP_PR_ELIM_UNUSED_VARS": None,
    "Z3_OP_PR_GOAL": None,
    "Z3_OP_PR_HYPER_RESOLVE": None,
    "Z3_OP_PR_HYPOTHESIS": None,
    "Z3_OP_PR_IFF_FALSE": None,
    "Z3_OP_PR_IFF_OEQ": None,
    "Z3_OP_PR_IFF_TRUE": None,
    "Z3_OP_PR_LEMMA": None,
    "Z3_OP_PR_LEMMA_ADD": None,
    "Z3_OP_PR_MODUS_PONENS": None,
    "Z3_OP_PR_MODUS_PONENS_OEQ": None,
    "Z3_OP_PR_MONOTONICITY": None,
    "Z3_OP_PR_NNF_NEG": None,
    "Z3_OP_PR_NNF_POS": None,
    "Z3_OP_PR_NOT_OR_ELIM": None,
    "Z3_OP_PR_PULL_QUANT": None,
    "Z3_OP_PR_PUSH_QUANT": None,
    "Z3_OP_PR_QUANT_INST": None,
    "Z3_OP_PR_QUANT_INTRO": None,
    "Z3_OP_PR_REDUNDANT_DEL": None,
    "Z3_OP_PR_REFLEXIVITY": None,
    "Z3_OP_PR_REWRITE": None,
    "Z3_OP_PR_REWRITE_STAR": None,
    "Z3_OP_PR_SKOLEMIZE": None,
    "Z3_OP_PR_SYMMETRY": None,
    "Z3_OP_PR_TH_LEMMA": None,
    "Z3_OP_PR_TRANSITIVITY": None,
    "Z3_OP_PR_TRANSITIVITY_STAR": None,
    "Z3_OP_PR_TRUE": None,
    "Z3_OP_PR_UNDEF": None,
    "Z3_OP_PR_UNIT_RESOLUTION": None,
    "Z3_OP_RA_CLONE": None,
    "Z3_OP_RA_COMPLEMENT": None,
    "Z3_OP_RA_EMPTY": None,
    "Z3_OP_RA_FILTER": None,
    "Z3_OP_RA_IS_EMPTY": None,
    "Z3_OP_RA_JOIN": None,
    "Z3_OP_RA_NEGATION_FILTER": None,
    "Z3_OP_RA_PROJECT": None,
    "Z3_OP_RA_RENAME": None,
    "Z3_OP_RA_SELECT": None,
    "Z3_OP_RA_STORE": None,
    "Z3_OP_RA_UNION": None,
    "Z3_OP_RA_WIDEN": None,
    "Z3_OP_RECURSIVE": None,
    "Z3_OP_REM": "__mod__",
    "Z3_OP_REPEAT": "RepeatBitVec",
    "Z3_OP_RE_COMPLEMENT": None,
    "Z3_OP_RE_CONCAT": None,
    "Z3_OP_RE_DERIVATIVE": None,
    "Z3_OP_RE_DIFF": None,
    "Z3_OP_RE_EMPTY_SET": None,
    "Z3_OP_RE_FULL_CHAR_SET": None,
    "Z3_OP_RE_FULL_SET": None,
    "Z3_OP_RE_INTERSECT": None,
    "Z3_OP_RE_LOOP": None,
    "Z3_OP_RE_OF_PRED": None,
    "Z3_OP_RE_OPTION": None,
    "Z3_OP_RE_PLUS": None,
    "Z3_OP_RE_POWER": None,
    "Z3_OP_RE_RANGE": None,
    "Z3_OP_RE_REVERSE": None,
    "Z3_OP_RE_STAR": None,
    "Z3_OP_RE_UNION": None,
    "Z3_OP_ROTATE_LEFT": None,
    "Z3_OP_ROTATE_RIGHT": None,
    "Z3_OP_SBV_TO_STR": None,
    "Z3_OP_SELECT": None,
    "Z3_OP_SEQ_AT": None,
    "Z3_OP_SEQ_CONCAT": None,
    "Z3_OP_SEQ_CONTAINS": None,
    "Z3_OP_SEQ_EMPTY": None,
    "Z3_OP_SEQ_EXTRACT": None,
    "Z3_OP_SEQ_INDEX": None,
    "Z3_OP_SEQ_IN_RE": None,
    "Z3_OP_SEQ_LAST_INDEX": None,
    "Z3_OP_SEQ_LENGTH": None,
    "Z3_OP_SEQ_NTH": None,
    "Z3_OP_SEQ_PREFIX": None,
    "Z3_OP_SEQ_REPLACE": None,
    "Z3_OP_SEQ_REPLACE_ALL": None,
    "Z3_OP_SEQ_REPLACE_RE": None,
    "Z3_OP_SEQ_REPLACE_RE_ALL": None,
    "Z3_OP_SEQ_SUFFIX": None,
    "Z3_OP_SEQ_TO_RE": None,
    "Z3_OP_SEQ_UNIT": None,
    "Z3_OP_SET_CARD": None,
    "Z3_OP_SET_COMPLEMENT": None,
    "Z3_OP_SET_DIFFERENCE": None,
    "Z3_OP_SET_HAS_SIZE": None,
    "Z3_OP_SET_INTERSECT": None,
    "Z3_OP_SET_SUBSET": None,
    "Z3_OP_SET_UNION": None,
    "Z3_OP_SGEQ": "SGE",
    "Z3_OP_SGT": "SGT",
    "Z3_OP_SIGN_EXT": "SignExt",
    "Z3_OP_SLEQ": "SLE",
    "Z3_OP_SLT": "SLT",
    "Z3_OP_SPECIAL_RELATION_LO": None,
    "Z3_OP_SPECIAL_RELATION_PLO": None,
    "Z3_OP_SPECIAL_RELATION_PO": None,
    "Z3_OP_SPECIAL_RELATION_TC": None,
    "Z3_OP_SPECIAL_RELATION_TO": None,
    "Z3_OP_SPECIAL_RELATION_TRC": None,
    "Z3_OP_STORE": None,
    "Z3_OP_STRING_LE": None,
    "Z3_OP_STRING_LT": None,
    "Z3_OP_STR_FROM_CODE": None,
    "Z3_OP_STR_TO_CODE": None,
    "Z3_OP_STR_TO_INT": None,
    "Z3_OP_SUB": "__sub__",
    "Z3_OP_TO_INT": None,
    "Z3_OP_TO_REAL": None,
    "Z3_OP_TRUE": "True",
    "Z3_OP_UBV_TO_STR": None,
    "Z3_OP_UGEQ": "UGE",
    "Z3_OP_UGT": "UGT",
    "Z3_OP_ULEQ": "ULE",
    "Z3_OP_ULT": "ULT",
    "Z3_OP_UMINUS": "__neg__",
    "Z3_OP_UNINTERPRETED": "UNINTERPRETED",
    "Z3_OP_XOR": "Xor",
    "Z3_OP_XOR3": None,
    "Z3_OP_ZERO_EXT": "ZeroExt",
}

op_type_map = {
    "Z3_OP_ADD": None,
    "Z3_OP_AGNUM": None,
    "Z3_OP_AND": Bool,
    "Z3_OP_ANUM": None,
    "Z3_OP_ARRAY_DEFAULT": None,
    "Z3_OP_ARRAY_EXT": None,
    "Z3_OP_ARRAY_MAP": None,
    "Z3_OP_AS_ARRAY": None,
    "Z3_OP_BADD": BV,
    "Z3_OP_BAND": BV,
    "Z3_OP_BASHR": BV,
    "Z3_OP_BCOMP": None,
    "Z3_OP_BIT0": None,
    "Z3_OP_BIT1": None,
    "Z3_OP_BIT2BOOL": None,
    "Z3_OP_BLSHR": BV,
    "Z3_OP_BMUL": BV,
    "Z3_OP_BNAND": None,
    "Z3_OP_BNEG": BV,
    "Z3_OP_BNOR": None,
    "Z3_OP_BNOT": BV,
    "Z3_OP_BNUM": "BitVecVal",
    "Z3_OP_BOR": BV,
    "Z3_OP_BREDAND": None,
    "Z3_OP_BREDOR": None,
    "Z3_OP_BSDIV": BV,
    "Z3_OP_BSDIV0": None,
    "Z3_OP_BSDIV_I": BV,
    "Z3_OP_BSHL": BV,
    "Z3_OP_BSMOD": BV,
    "Z3_OP_BSMOD0": None,
    "Z3_OP_BSMOD_I": BV,
    "Z3_OP_BSMUL_NO_OVFL": None,
    "Z3_OP_BSMUL_NO_UDFL": None,
    "Z3_OP_BSREM": BV,
    "Z3_OP_BSREM0": None,
    "Z3_OP_BSREM_I": BV,
    "Z3_OP_BSUB": BV,
    "Z3_OP_BUDIV": BV,
    "Z3_OP_BUDIV0": None,
    "Z3_OP_BUDIV_I": BV,
    "Z3_OP_BUMUL_NO_OVFL": None,
    "Z3_OP_BUREM": BV,
    "Z3_OP_BUREM0": None,
    "Z3_OP_BUREM_I": BV,
    "Z3_OP_BV2INT": None,
    "Z3_OP_BXNOR": None,
    "Z3_OP_BXOR": BV,
    "Z3_OP_CARRY": None,
    "Z3_OP_CHAR_CONST": None,
    "Z3_OP_CHAR_FROM_BV": None,
    "Z3_OP_CHAR_IS_DIGIT": None,
    "Z3_OP_CHAR_LE": None,
    "Z3_OP_CHAR_TO_BV": None,
    "Z3_OP_CHAR_TO_INT": None,
    "Z3_OP_CONCAT": BV,
    "Z3_OP_CONST_ARRAY": None,
    "Z3_OP_DISTINCT": Bool,
    "Z3_OP_DIV": None,
    "Z3_OP_DT_ACCESSOR": None,
    "Z3_OP_DT_CONSTRUCTOR": None,
    "Z3_OP_DT_IS": None,
    "Z3_OP_DT_RECOGNISER": None,
    "Z3_OP_DT_UPDATE_FIELD": None,
    "Z3_OP_EQ": Bool,
    "Z3_OP_EXTRACT": BV,
    "Z3_OP_EXT_ROTATE_LEFT": BV,
    "Z3_OP_EXT_ROTATE_RIGHT": BV,
    "Z3_OP_FALSE": Bool,
    "Z3_OP_FD_CONSTANT": None,
    "Z3_OP_FD_LT": None,
    "Z3_OP_FPA_ABS": FP,
    "Z3_OP_FPA_ADD": FP,
    "Z3_OP_FPA_BV2RM": None,
    "Z3_OP_FPA_BVWRAP": None,
    "Z3_OP_FPA_DIV": FP,
    "Z3_OP_FPA_EQ": Bool,
    "Z3_OP_FPA_FMA": None,
    "Z3_OP_FPA_FP": None,
    "Z3_OP_FPA_GE": Bool,
    "Z3_OP_FPA_GT": Bool,
    "Z3_OP_FPA_IS_INF": None,
    "Z3_OP_FPA_IS_NAN": None,
    "Z3_OP_FPA_IS_NEGATIVE": None,
    "Z3_OP_FPA_IS_NORMAL": None,
    "Z3_OP_FPA_IS_POSITIVE": None,
    "Z3_OP_FPA_IS_SUBNORMAL": None,
    "Z3_OP_FPA_IS_ZERO": None,
    "Z3_OP_FPA_LE": Bool,
    "Z3_OP_FPA_LT": Bool,
    "Z3_OP_FPA_MAX": None,
    "Z3_OP_FPA_MIN": None,
    "Z3_OP_FPA_MINUS_INF": FP,
    "Z3_OP_FPA_MINUS_ZERO": FP,
    "Z3_OP_FPA_MUL": FP,
    "Z3_OP_FPA_NAN": FP,
    "Z3_OP_FPA_NEG": FP,
    "Z3_OP_FPA_NUM": FP,
    "Z3_OP_FPA_PLUS_INF": FP,
    "Z3_OP_FPA_PLUS_ZERO": FP,
    "Z3_OP_FPA_REM": None,
    "Z3_OP_FPA_RM_NEAREST_TIES_TO_AWAY": None,
    "Z3_OP_FPA_RM_NEAREST_TIES_TO_EVEN": None,
    "Z3_OP_FPA_RM_TOWARD_NEGATIVE": None,
    "Z3_OP_FPA_RM_TOWARD_POSITIVE": None,
    "Z3_OP_FPA_RM_TOWARD_ZERO": None,
    "Z3_OP_FPA_ROUND_TO_INTEGRAL": None,
    "Z3_OP_FPA_SQRT": FP,
    "Z3_OP_FPA_SUB": FP,
    "Z3_OP_FPA_TO_FP": FP,
    "Z3_OP_FPA_TO_FP_UNSIGNED": FP,
    "Z3_OP_FPA_TO_IEEE_BV": BV,
    "Z3_OP_FPA_TO_REAL": None,
    "Z3_OP_FPA_TO_SBV": BV,
    "Z3_OP_FPA_TO_UBV": BV,
    "Z3_OP_GE": None,
    "Z3_OP_GT": None,
    "Z3_OP_IDIV": None,
    "Z3_OP_IFF": Bool,
    "Z3_OP_IMPLIES": None,
    "Z3_OP_INT2BV": None,
    "Z3_OP_INTERNAL": None,
    "Z3_OP_INT_TO_STR": None,
    "Z3_OP_IS_INT": None,
    "Z3_OP_ITE": Bool,
    "Z3_OP_LABEL": None,
    "Z3_OP_LABEL_LIT": None,
    "Z3_OP_LE": None,
    "Z3_OP_LT": None,
    "Z3_OP_MOD": None,
    "Z3_OP_MUL": None,
    "Z3_OP_NOT": Bool,
    "Z3_OP_OEQ": None,
    "Z3_OP_OR": Bool,
    "Z3_OP_PB_AT_LEAST": None,
    "Z3_OP_PB_AT_MOST": None,
    "Z3_OP_PB_EQ": None,
    "Z3_OP_PB_GE": None,
    "Z3_OP_PB_LE": None,
    "Z3_OP_POWER": None,
    "Z3_OP_PR_AND_ELIM": None,
    "Z3_OP_PR_APPLY_DEF": None,
    "Z3_OP_PR_ASSERTED": None,
    "Z3_OP_PR_ASSUMPTION_ADD": None,
    "Z3_OP_PR_BIND": None,
    "Z3_OP_PR_CLAUSE_TRAIL": None,
    "Z3_OP_PR_COMMUTATIVITY": None,
    "Z3_OP_PR_DEF_AXIOM": None,
    "Z3_OP_PR_DEF_INTRO": None,
    "Z3_OP_PR_DER": None,
    "Z3_OP_PR_DISTRIBUTIVITY": None,
    "Z3_OP_PR_ELIM_UNUSED_VARS": None,
    "Z3_OP_PR_GOAL": None,
    "Z3_OP_PR_HYPER_RESOLVE": None,
    "Z3_OP_PR_HYPOTHESIS": None,
    "Z3_OP_PR_IFF_FALSE": None,
    "Z3_OP_PR_IFF_OEQ": None,
    "Z3_OP_PR_IFF_TRUE": None,
    "Z3_OP_PR_LEMMA": None,
    "Z3_OP_PR_LEMMA_ADD": None,
    "Z3_OP_PR_MODUS_PONENS": None,
    "Z3_OP_PR_MODUS_PONENS_OEQ": None,
    "Z3_OP_PR_MONOTONICITY": None,
    "Z3_OP_PR_NNF_NEG": None,
    "Z3_OP_PR_NNF_POS": None,
    "Z3_OP_PR_NOT_OR_ELIM": None,
    "Z3_OP_PR_PULL_QUANT": None,
    "Z3_OP_PR_PUSH_QUANT": None,
    "Z3_OP_PR_QUANT_INST": None,
    "Z3_OP_PR_QUANT_INTRO": None,
    "Z3_OP_PR_REDUNDANT_DEL": None,
    "Z3_OP_PR_REFLEXIVITY": None,
    "Z3_OP_PR_REWRITE": None,
    "Z3_OP_PR_REWRITE_STAR": None,
    "Z3_OP_PR_SKOLEMIZE": None,
    "Z3_OP_PR_SYMMETRY": None,
    "Z3_OP_PR_TH_LEMMA": None,
    "Z3_OP_PR_TRANSITIVITY": None,
    "Z3_OP_PR_TRANSITIVITY_STAR": None,
    "Z3_OP_PR_TRUE": None,
    "Z3_OP_PR_UNDEF": None,
    "Z3_OP_PR_UNIT_RESOLUTION": None,
    "Z3_OP_RA_CLONE": None,
    "Z3_OP_RA_COMPLEMENT": None,
    "Z3_OP_RA_EMPTY": None,
    "Z3_OP_RA_FILTER": None,
    "Z3_OP_RA_IS_EMPTY": None,
    "Z3_OP_RA_JOIN": None,
    "Z3_OP_RA_NEGATION_FILTER": None,
    "Z3_OP_RA_PROJECT": None,
    "Z3_OP_RA_RENAME": None,
    "Z3_OP_RA_SELECT": None,
    "Z3_OP_RA_STORE": None,
    "Z3_OP_RA_UNION": None,
    "Z3_OP_RA_WIDEN": None,
    "Z3_OP_RECURSIVE": None,
    "Z3_OP_REM": None,
    "Z3_OP_REPEAT": BV,
    "Z3_OP_RE_COMPLEMENT": None,
    "Z3_OP_RE_CONCAT": None,
    "Z3_OP_RE_DERIVATIVE": None,
    "Z3_OP_RE_DIFF": None,
    "Z3_OP_RE_EMPTY_SET": None,
    "Z3_OP_RE_FULL_CHAR_SET": None,
    "Z3_OP_RE_FULL_SET": None,
    "Z3_OP_RE_INTERSECT": None,
    "Z3_OP_RE_LOOP": None,
    "Z3_OP_RE_OF_PRED": None,
    "Z3_OP_RE_OPTION": None,
    "Z3_OP_RE_PLUS": None,
    "Z3_OP_RE_POWER": None,
    "Z3_OP_RE_RANGE": None,
    "Z3_OP_RE_REVERSE": None,
    "Z3_OP_RE_STAR": None,
    "Z3_OP_RE_UNION": None,
    "Z3_OP_ROTATE_LEFT": None,
    "Z3_OP_ROTATE_RIGHT": None,
    "Z3_OP_SBV_TO_STR": None,
    "Z3_OP_SELECT": None,
    "Z3_OP_SEQ_AT": None,
    "Z3_OP_SEQ_CONCAT": None,
    "Z3_OP_SEQ_CONTAINS": None,
    "Z3_OP_SEQ_EMPTY": None,
    "Z3_OP_SEQ_EXTRACT": None,
    "Z3_OP_SEQ_INDEX": None,
    "Z3_OP_SEQ_IN_RE": None,
    "Z3_OP_SEQ_LAST_INDEX": None,
    "Z3_OP_SEQ_LENGTH": None,
    "Z3_OP_SEQ_NTH": None,
    "Z3_OP_SEQ_PREFIX": None,
    "Z3_OP_SEQ_REPLACE": None,
    "Z3_OP_SEQ_REPLACE_ALL": None,
    "Z3_OP_SEQ_REPLACE_RE": None,
    "Z3_OP_SEQ_REPLACE_RE_ALL": None,
    "Z3_OP_SEQ_SUFFIX": None,
    "Z3_OP_SEQ_TO_RE": None,
    "Z3_OP_SEQ_UNIT": None,
    "Z3_OP_SET_CARD": None,
    "Z3_OP_SET_COMPLEMENT": None,
    "Z3_OP_SET_DIFFERENCE": None,
    "Z3_OP_SET_HAS_SIZE": None,
    "Z3_OP_SET_INTERSECT": None,
    "Z3_OP_SET_SUBSET": None,
    "Z3_OP_SET_UNION": None,
    "Z3_OP_SGEQ": Bool,
    "Z3_OP_SGT": Bool,
    "Z3_OP_SIGN_EXT": BV,
    "Z3_OP_SLEQ": Bool,
    "Z3_OP_SLT": Bool,
    "Z3_OP_SPECIAL_RELATION_LO": None,
    "Z3_OP_SPECIAL_RELATION_PLO": None,
    "Z3_OP_SPECIAL_RELATION_PO": None,
    "Z3_OP_SPECIAL_RELATION_TC": None,
    "Z3_OP_SPECIAL_RELATION_TO": None,
    "Z3_OP_SPECIAL_RELATION_TRC": None,
    "Z3_OP_STORE": None,
    "Z3_OP_STRING_LE": None,
    "Z3_OP_STRING_LT": None,
    "Z3_OP_STR_FROM_CODE": None,
    "Z3_OP_STR_TO_CODE": None,
    "Z3_OP_STR_TO_INT": None,
    "Z3_OP_SUB": None,
    "Z3_OP_TO_INT": None,
    "Z3_OP_TO_REAL": None,
    "Z3_OP_TRUE": Bool,
    "Z3_OP_UBV_TO_STR": None,
    "Z3_OP_UGEQ": Bool,
    "Z3_OP_UGT": Bool,
    "Z3_OP_ULEQ": Bool,
    "Z3_OP_ULT": Bool,
    "Z3_OP_UMINUS": None,
    "Z3_OP_UNINTERPRETED": "UNINTERPRETED",
    "Z3_OP_XOR": Bool,
    "Z3_OP_XOR3": None,
    "Z3_OP_ZERO_EXT": BV,
}



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

import claripy


class TestFallbackAbstraction(unittest.TestCase):
    def test_fallback_abstraction(self):
        bz = claripy.backends.z3

        a = claripy.BVV(5, 32)
        b = claripy.BVS("x", 32, explicit_name=True)
        c = a + b
        d = 5 + b
        e = b + 5
        f = b + b
        g = a + a

        assert not a.symbolic
        assert b.symbolic
        assert c.symbolic
        assert d.symbolic
        assert e.symbolic
        assert f.symbolic

        assert type(claripy.backends.concrete.convert(a)) is claripy.backends.backend_concrete.BVV
        assert type(claripy.backends.concrete.convert(g)) is claripy.backends.backend_concrete.BVV

        self.assertRaises(claripy.errors.BackendError, claripy.backends.concrete.convert, b)
        self.assertRaises(claripy.errors.BackendError, claripy.backends.concrete.convert, c)
        self.assertRaises(claripy.errors.BackendError, claripy.backends.concrete.convert, d)
        self.assertRaises(claripy.errors.BackendError, claripy.backends.concrete.convert, e)
        self.assertRaises(claripy.errors.BackendError, claripy.backends.concrete.convert, f)

        assert str(bz.convert(b)) == "x"
        assert bz.convert(b).__module__ == "z3.z3"

        assert str(bz.convert(c)) == "5 + x"
        assert str(bz.convert(d)) == "5 + x"
        assert str(bz.convert(e)) == "x + 5"
        assert str(bz.convert(f)) == "x + x"


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from claripy.ast import Base
from claripy.errors import ClaripyReplacementError

if TYPE_CHECKING:
    from collections.abc import Callable

T = TypeVar("T", bound="Base")


def replace_dict(
    expr: Base,
    replacements: dict[int, Base],
    variable_set: set[str] | None = None,
    leaf_operation: Callable[[Base], Base] = lambda x: x,
) -> Base:
    """
    Returns this AST with subexpressions replaced by those that can be found in `replacements`
    dict.

    :param variable_set:        For optimization, ast's without these variables are not checked
                                    for replacing.
    :param replacements:        A dictionary of hashes to their replacements.
    :param leaf_operation:      An operation that should be applied to the leaf nodes.
    :returns:                   An AST with all instances of ast's in replacements.
    """
    if variable_set is None:
        variable_set = set()

    arg_queue = [iter([expr])]
    rep_queue = []
    ast_queue = []

    while arg_queue:
        try:
            ast = next(arg_queue[-1])
            repl = ast

            if not isinstance(ast, Base):
                rep_queue.append(repl)
                continue

            if ast.hash() in replacements:
                repl = replacements[ast.hash()]

            elif ast.variables >= variable_set:
                if ast.is_leaf():
                    repl = leaf_operation(ast)
                    if repl is not ast:
                        replacements[ast.hash()] = repl

                elif ast.depth > 1:
                    arg_queue.append(iter(ast.args))
                    ast_queue.append(ast)
                    continue

            rep_queue.append(repl)

        except StopIteration:
            arg_queue.pop()

            if ast_queue:
                ast = ast_queue.pop()
                repl = ast

                args = rep_queue[-len(ast.args) :]
                del rep_queue[-len(ast.args) :]

                # Check if replacement occurred.
                if any((a is not b for a, b in zip(ast.args, args, strict=False))):
                    repl = ast.make_like(ast.op, tuple(args))
                    replacements[ast.hash()] = repl

                rep_queue.append(repl)

    assert len(arg_queue) == 0, "arg_queue is not empty"
    assert len(ast_queue) == 0, "ast_queue is not empty"
    assert len(rep_queue) == 1, ("rep_queue has unexpected length", len(rep_queue))

    return rep_queue.pop()


def _check_replaceability(old: T, new: T) -> None:
    if not isinstance(old, Base) or not isinstance(new, Base):
        raise ClaripyReplacementError("replacements must be AST nodes")
    if type(old) is not type(new):
        raise ClaripyReplacementError(f"cannot replace type {type(old)} ast with type {type(new)} ast")


def replace(expr: Base, old: T, new: T) -> Base:
    """
    Returns this AST but with the AST 'old' replaced with AST 'new' in its subexpressions.
    """
    _check_replaceability(old, new)
    replacements: dict[int, Base] = {old.hash(): new}
    return replace_dict(expr, replacements, variable_set=old.variables)



from __future__ import annotations

import functools
import logging
import numbers
import operator
from functools import reduce

import claripy
from claripy.annotation import RegionAnnotation, StridedIntervalAnnotation, UninitializedAnnotation
from claripy.ast import BV, Base
from claripy.backends.backend import Backend
from claripy.backends.backend_vsa.balancer import Balancer
from claripy.backends.backend_vsa.errors import ClaripyVSAError
from claripy.errors import BackendError
from claripy.operations import backend_operations_vsa_compliant, expression_set_operations

from .bool_result import BoolResult, FalseResult, TrueResult
from .discrete_strided_interval_set import DiscreteStridedIntervalSet
from .strided_interval import StridedInterval
from .valueset import ValueSet

log = logging.getLogger(__name__)


def arg_filter(f):
    @functools.wraps(f)
    def filter_(*args):
        if isinstance(args[0], numbers.Number):
            raise BackendError(f"Unsupported argument type {type(args[0])}")
        return f(*args)

    return filter_


def normalize_arg_order(f):
    @functools.wraps(f)
    def normalizer(*args):
        if len(args) != 2:
            raise BackendError(f"Unsupported arguments number {len(args)}")

        if not isinstance(args[0], StridedInterval | DiscreteStridedIntervalSet | ValueSet):
            if not isinstance(args[1], StridedInterval | DiscreteStridedIntervalSet | ValueSet):
                raise BackendError("Unsupported arguments")
            args = [args[1], args[0]]

        return f(*args)

    return normalizer


# pylint: disable=too-many-positional-arguments


class BackendVSA(Backend):
    """BackendVSA is a backend that uses VSA (Value Set Analysis) to represent
    and reason about values.
    """

    def __init__(self):
        Backend.__init__(self)
        self._make_expr_ops(set(expression_set_operations), op_class=self)
        self._make_raw_ops(set(backend_operations_vsa_compliant), op_module=BackendVSA)

        self._op_raw["Reverse"] = BackendVSA.Reverse
        self._op_raw["If"] = self.If
        self._op_expr["BVV"] = self.BVV
        self._op_expr["BoolV"] = self.BoolV
        self._op_expr["BVS"] = self.BVS

        # reduceable
        self._op_raw["__add__"] = self._op_add
        self._op_raw["__sub__"] = self._op_sub
        self._op_raw["__mul__"] = self._op_mul
        self._op_raw["__or__"] = self._op_or
        self._op_raw["__xor__"] = self._op_xor
        self._op_raw["__and__"] = self._op_and
        self._op_raw["__mod__"] = self._op_mod

    @staticmethod
    def _op_add(*args):
        return reduce(operator.__add__, args)

    @staticmethod
    def _op_sub(*args):
        return reduce(operator.__sub__, args)

    @staticmethod
    def _op_mul(*args):
        return reduce(operator.__mul__, args)

    @staticmethod
    def _op_or(*args):
        return reduce(operator.__or__, args)

    @staticmethod
    def _op_xor(*args):
        return reduce(operator.__xor__, args)

    @staticmethod
    def _op_and(*args):
        return reduce(operator.__and__, args)

    @staticmethod
    def _op_mod(*args):
        return reduce(operator.__mod__, args)

    def convert(self, expr):
        return Backend.convert(self, claripy.excavate_ite(expr) if isinstance(expr, Base) else expr)

    def _convert(self, r):
        if isinstance(r, numbers.Number):
            return r
        if isinstance(r, bool):
            return TrueResult() if r else FalseResult()
        if isinstance(r, StridedInterval | DiscreteStridedIntervalSet | ValueSet):
            return r
        if isinstance(r, BoolResult):
            return r

        # Not supported
        raise BackendError

    def _abstract(self, e):
        if isinstance(e, numbers.Number):
            return e
        if isinstance(e, StridedInterval):
            if e.is_top:
                return claripy.TSI(e.bits, explicit_name=e.name)
            if e.is_empty:
                return claripy.ESI(e.bits)
            if e.stride in {0, 1} and e.lower_bound == e.upper_bound:
                return claripy.BVV(e.lower_bound, e.bits)
            return claripy.SI(
                name=e.name,
                bits=e.bits,
                lower_bound=e.lower_bound,
                upper_bound=e.upper_bound,
                stride=e.stride,
            )
        if isinstance(e, ValueSet):
            if len(e.regions) == 0:
                return claripy.VS(bits=e.bits, name=e.name)
            if len(e.regions) == 1:
                region = next(iter(e.regions))
                return claripy.VS(
                    bits=e.bits,
                    region=region,
                    region_base_addr=e._region_base_addrs[region].eval(1)[0] if e._region_base_addrs else 0,
                    value=e.regions[region].eval(1)[0],
                    name=e.name,
                )
            raise ClaripyVSAError("Cannot abstract ValueSet with multiple regions")
        if isinstance(e, BoolResult):
            if e.is_true:
                return claripy.BoolV(True)
            if e.is_false:
                return claripy.BoolV(False)
            return claripy.BoolS("maybe")
        raise BackendError(f"Don't know how to abstract {type(e)}")

    def _eval(self, expr, n, extra_constraints=(), solver=None, model_callback=None):
        if isinstance(expr, StridedInterval | ValueSet):
            return expr.eval(n)
        if isinstance(expr, BoolResult):
            return expr.value
        raise BackendError(f"Unsupported type {type(expr)}")

    def _min(self, expr, extra_constraints=(), signed=False, solver=None, model_callback=None):
        if isinstance(expr, StridedInterval | ValueSet):
            return expr.min(signed=signed)

        raise BackendError(f"Unsupported expr type {type(expr)}")

    def _max(self, expr, extra_constraints=(), signed=False, solver=None, model_callback=None):
        if isinstance(expr, StridedInterval | ValueSet):
            return expr.max(signed=signed)

        raise BackendError(f"Unsupported expr type {type(expr)}")

    def _solution(self, expr, v, extra_constraints=(), solver=None, model_callback=None):
        if isinstance(expr, BoolResult):
            return len(set(v.value) & set(expr.value)) > 0

        if isinstance(expr, StridedInterval):
            return not expr.intersection(v).is_empty

        if isinstance(expr, ValueSet):
            return any(not si.intersection(v).is_empty for _, si in expr.items())

        raise NotImplementedError(type(expr).__name__)

    def _has_true(self, e, extra_constraints=(), solver=None, model_callback=None):
        return BoolResult.has_true(e)

    def _has_false(self, e, extra_constraints=(), solver=None, model_callback=None):
        return BoolResult.has_false(e)

    def _is_true(self, e, extra_constraints=(), solver=None, model_callback=None):
        return BoolResult.is_true(e)

    def _is_false(self, e, extra_constraints=(), solver=None, model_callback=None):
        return BoolResult.is_false(e)

    #
    # Backend Operations
    #

    def _identical(self, a, b):
        if type(a) != type(b):  # noqa: E721
            return False
        return a.identical(b)

    def _cardinality(self, a):
        return a.cardinality

    def name(self, a):
        if isinstance(a, StridedInterval):
            return a.name

        return None

    def apply_annotation(self, o, a):
        """
        Apply an annotation on the backend object.

        :param BackendObject bo: The backend object.
        :param Annotation annotation: The annotation to be applied
        :return: A new BackendObject
        :rtype: BackendObject
        """

        if isinstance(o, StridedInterval):
            if isinstance(a, StridedIntervalAnnotation):
                return StridedInterval(
                    bits=o.bits,
                    stride=a.stride,
                    lower_bound=a.lower_bound,
                    upper_bound=a.upper_bound,
                    name=o.name,
                )

            if isinstance(a, RegionAnnotation):
                offset = o
                if isinstance(offset, numbers.Number):
                    offset = StridedInterval(bits=o.bits, stride=0, lower_bound=offset, upper_bound=offset)
                vs = ValueSet.empty(o.bits)
                if isinstance(offset, StridedInterval):
                    vs._merge_si(a.region_id, a.region_base_addr, offset)
                elif isinstance(offset, ValueSet):
                    for si in offset.regions.values():
                        vs._merge_si(a.region_id, a.region_base_addr, si)
                else:
                    raise ClaripyVSAError(f"Unsupported offset type {type(offset)}")
                return vs

            if isinstance(a, UninitializedAnnotation):
                o2 = o.copy()
                o2.uninitialized = True
                return o2

        if isinstance(o, ValueSet) and isinstance(a, StridedIntervalAnnotation):
            si = StridedInterval(
                bits=o.bits,
                stride=a.stride,
                lower_bound=a.lower_bound,
                upper_bound=a.upper_bound,
                name=o.name,
            )
            vs = o.copy()
            vs._merge_si(a.region_id, a.region_base_addr, si)
            return vs

        if isinstance(o, BoolResult) and isinstance(a, UninitializedAnnotation):
            # TODO: Do we want to do anything here?
            return o

        raise ValueError(f"Unsupported annotation type {type(a)} for object {type(o)}")

    @staticmethod
    def BVV(ast):
        if ast.args[0] is None:
            return StridedInterval.empty(ast.args[1])
        return StridedInterval(bits=ast.args[1], stride=0, lower_bound=ast.args[0], upper_bound=ast.args[0])

    @staticmethod
    def BoolV(ast):
        return TrueResult() if ast.args[0] else FalseResult()

    @staticmethod
    def And(a, *args):
        return reduce(operator.__and__, args, a)

    @staticmethod
    def Not(a):
        return ~a

    @staticmethod
    @normalize_arg_order
    def ULT(a, b):
        return a.ULT(b)

    @staticmethod
    @normalize_arg_order
    def ULE(a, b):
        return a.ULE(b)

    @staticmethod
    @normalize_arg_order
    def UGT(a, b):
        return a.UGT(b)

    @staticmethod
    @normalize_arg_order
    def UGE(a, b):
        return a.UGE(b)

    @staticmethod
    @normalize_arg_order
    def SLT(a, b):
        return a.SLT(b)

    @staticmethod
    @normalize_arg_order
    def SLE(a, b):
        return a.SLE(b)

    @staticmethod
    @normalize_arg_order
    def SGT(a, b):
        return a.SGT(b)

    @staticmethod
    @normalize_arg_order
    def SGE(a, b):
        return a.SGE(b)

    @staticmethod
    def BVS(ast: BV):
        return StridedInterval(name=ast.args[0], bits=ast.size())

    def If(self, cond, t, f):
        if not self.has_true(cond):
            return f
        if not self.has_false(cond):
            return t
        return t.union(f)

    # TODO: Implement other operations!

    @staticmethod
    def Or(*args):
        first = args[0]
        others = args[1:]

        for o in others:
            first = first.union(o)
        return first

    @staticmethod
    def __rshift__(expr, shift_amount):  # pylint:disable=unexpected-special-method-signature
        return expr.__rshift__(shift_amount)

    @staticmethod
    def LShR(expr, shift_amount):
        return expr.LShR(shift_amount)

    @staticmethod
    def Concat(*args):
        ret = None
        for expr in args:
            if not isinstance(expr, StridedInterval | DiscreteStridedIntervalSet | ValueSet):
                raise BackendError(f"Unsupported expr type {type(expr)}")

            ret = ret.concat(expr) if ret is not None else expr

        return ret

    @staticmethod
    def Extract(*args):
        low_bit = args[1]
        high_bit = args[0]
        expr = args[2]

        if not isinstance(expr, StridedInterval | DiscreteStridedIntervalSet | ValueSet):
            raise BackendError(f"Unsupported expr type {type(expr)}")

        return expr.extract(high_bit, low_bit)

    @staticmethod
    def SignExt(*args):
        new_bits = args[0]
        expr = args[1]

        if not isinstance(expr, StridedInterval | DiscreteStridedIntervalSet):
            raise BackendError(f"Unsupported expr type {type(expr)}")

        return expr.sign_extend(new_bits + expr.bits)

    @staticmethod
    def ZeroExt(*args):
        new_bits = args[0]
        expr = args[1]

        if not isinstance(expr, StridedInterval | DiscreteStridedIntervalSet):
            raise BackendError(f"Unsupported expr type {type(expr)}")

        return expr.zero_extend(new_bits + expr.bits)

    @staticmethod
    def Reverse(arg):
        if not isinstance(arg, StridedInterval | DiscreteStridedIntervalSet | ValueSet):
            raise BackendError(f"Unsupported expr type {type(arg)}")

        return arg.reverse()

    def union(self, ast):
        if len(ast.args) != 2:
            raise BackendError(f"Incorrect number of arguments ({len(ast.args)}) passed to BackendVSA.union().")

        converted_0 = self.convert(ast.args[0])
        converted_1 = self.convert(ast.args[1])

        ret = converted_0.union(converted_1)

        if ret is NotImplemented:
            log.debug("Union failed, trying the other way around.")
            ret = converted_1.union(converted_0)

        return ret

    def intersection(self, ast):
        if len(ast.args) != 2:
            raise BackendError(f"Incorrect number of arguments ({len(ast.args)}) passed to BackendVSA.intersection().")

        ret = None

        for arg in ast.args:
            arg = self.convert(arg)
            ret = arg if ret is None else ret.intersection(arg)
        return ret

    def widen(self, ast):
        if len(ast.args) != 2:
            raise BackendError(f"Incorrect number of arguments ({len(ast.args)}) passed to BackendVSA.widen().")

        converted_0 = self.convert(ast.args[0])
        converted_1 = self.convert(ast.args[1])

        ret = converted_0.widen(converted_1)
        if ret is NotImplemented:
            log.debug("Widening failed, trying the other way around.")
            ret = converted_1.widen(converted_0)

        return ret

    @staticmethod
    def CreateTopStridedInterval(bits, name=None):
        return StridedInterval.top(bits, name)

    @staticmethod
    def constraint_to_si(expr):
        return Balancer(expr).compat_ret



from __future__ import annotations

from typing import cast

from claripy.ast.base import Base
from claripy.errors import ClaripyOperationError


class Bits(Base):
    """
    A base class for AST types that can be stored as a series of bits. Currently, this is bitvectors and IEEE floats.

    :ivar length:       The length of this value in bits.
    """

    __slots__ = ()

    def make_like(self, op, args, **kwargs):
        if "length" not in kwargs:
            kwargs["length"] = self.length
        return Base.make_like(self, op, args, **kwargs)

    def size(self) -> int:
        """
        :returns: The bit length of this AST
        """
        return cast("int", self.length)

    def _type_name(self):
        return self.__class__.__name__ + str(self.length)

    def raw_to_bv(self):
        """
        Converts this data's bit-pattern to a bitvector.
        """
        raise NotImplementedError

    def raw_to_fp(self):
        """
        Converts this data's bit-pattern to an IEEE float.
        """
        raise NotImplementedError

    @staticmethod
    def _check_replaceability(old, new):
        if old.size() != new.size():
            raise ClaripyOperationError("replacements must have matching sizes")

    __len__ = size



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, TypeVar

import claripy
import claripy.simplifications

from . import debug as _d
from .errors import ClaripyOperationError, ClaripyTypeError

if TYPE_CHECKING:
    from collections.abc import Callable

    from claripy.ast import Base

T = TypeVar("T", bound="Base")


def op(name, arg_types, return_type: type[T], extra_check=None, calc_length=None) -> Callable[..., T]:
    if isinstance(arg_types, tuple | list):
        expected_num_args = len(arg_types)
    elif isinstance(arg_types, type):
        expected_num_args = None
    else:
        raise ClaripyOperationError(f"op {name} got weird arg_types")

    def _type_fixer(args):
        num_args = len(args)
        if expected_num_args is not None and num_args != expected_num_args:
            if num_args + 1 == expected_num_args and arg_types[0] is claripy.fp.RM:
                args = (claripy.fp.RM.default(), *args)
            else:
                raise ClaripyTypeError(f"Operation {name} takes exactly {len(arg_types)} arguments ({len(args)} given)")

        actual_arg_types = (arg_types,) * num_args if isinstance(arg_types, type) else arg_types
        matches = list(itertools.starmap(isinstance, zip(args, actual_arg_types, strict=False)))

        # heuristically, this works!
        thing = args[matches.index(True, 1 if actual_arg_types[0] is claripy.fp.RM else 0)] if True in matches else None

        for arg, argty, match in zip(args, actual_arg_types, matches, strict=False):
            if not match:
                if hasattr(argty, "_from_" + type(arg).__name__):
                    convert = getattr(argty, "_from_" + type(arg).__name__)
                    yield convert(thing, arg)
                else:
                    yield NotImplemented
                    return
            else:
                yield arg

    def _op(*args):
        fixed_args = tuple(_type_fixer(args))
        if _d._DEBUG:
            for i in fixed_args:
                if i is NotImplemented:
                    return NotImplemented
            if extra_check is not None:
                success, msg = extra_check(*fixed_args)
                if not success:
                    raise ClaripyOperationError(msg)

        # pylint:disable=too-many-nested-blocks
        simp, annotated = claripy.simplifications.simplify(name, fixed_args)
        if simp is not None:
            if not annotated:
                simp = _handle_annotations(simp, fixed_args)
            if simp is not None:
                return simp

        kwargs = {}
        if calc_length is not None:
            kwargs["length"] = calc_length(*fixed_args)

        return return_type(name, fixed_args, **kwargs)

    _op.calc_length = calc_length
    return _op


def _handle_annotations(simp, args):
    # pylint:disable=isinstance-second-argument-not-valid-type
    ast_args = tuple(a for a in args if isinstance(a, claripy.ast.Base))
    preserved_relocatable = frozenset(simp._relocatable_annotations)
    relocated_annotations = set()
    bad_eliminated = 0

    for aa in ast_args:
        for oa in aa._relocatable_annotations:
            if oa not in preserved_relocatable and oa not in relocated_annotations:
                relocated_annotations.add(oa)
                na = oa.relocate(aa, simp)
                if na is not None:
                    simp = simp.append_annotation(na)

        bad_eliminated += len(aa._uneliminatable_annotations - simp._uneliminatable_annotations)

    if bad_eliminated == 0:
        return simp
    return None


def reversed_op(op_func):
    if type(op_func) is not type(reversed_op):
        op_func = op_func.im_func  # unwrap instancemethod into function

    def _reversed_op(*args):
        return op_func(*args[::-1])

    return _reversed_op


#
# Length checkers
#


def length_same_check(*args):
    return all(a.length == args[0].length for a in args), "args' length must all be equal"


def basic_length_calc(*args):
    return args[0].length


def extract_check(high, low, bv):
    if high < 0 or low < 0:
        return False, "Extract high and low must be nonnegative"
    if low > high:
        return False, "Extract low must be <= high"
    if high >= bv.size():
        return False, "Extract bound must be less than BV size"

    return True, ""


def extend_check(amount, _):
    return amount >= 0, "Extension length must be nonnegative"


def concat_length_calc(*args):
    return sum(arg.length for arg in args)


def extract_length_calc(high, low, _):
    return high + 1 - low


def ext_length_calc(ext, orig):
    return orig.length + ext


#
# Operation lists
#

expression_arithmetic_operations = {
    # arithmetic
    "__add__",
    "__radd__",
    "__truediv__",
    "__rtruediv__",
    "__floordiv__",
    "__rfloordiv__",
    "__mul__",
    "__rmul__",
    "__sub__",
    "__rsub__",
    "__mod__",
    "__rmod__",
    "SDiv",
    "SMod",
    "__neg__",
    "__abs__",
}

bin_ops = {
    "__add__",
    "__radd__",
    "__mul__",
    "__rmul__",
    "__or__",
    "__ror__",
    "__and__",
    "__rand__",
    "__xor__",
    "__rxor__",
}

expression_comparator_operations = {
    # comparisons
    "__eq__",
    "__ne__",
    "UGT",
    "UGE",
    "ULT",
    "ULE",
    "SGT",
    "SGE",
    "SLT",
    "SLE",
}

expression_bitwise_operations = {
    # bitwise
    "__invert__",
    "__or__",
    "__ror__",
    "__and__",
    "__rand__",
    "__xor__",
    "__rxor__",
    "__lshift__",
    "__rlshift__",
    "__rshift__",
    "__rrshift__",
}

expression_set_operations = {
    # Set operations
    "union",
    "intersection",
    "widen",
}

expression_operations = (
    expression_arithmetic_operations
    | expression_comparator_operations
    | expression_bitwise_operations
    | expression_set_operations
)

backend_comparator_operations = {
    "SGE",
    "SLE",
    "SGT",
    "SLT",
    "UGE",
    "ULE",
    "UGT",
    "ULT",
}

backend_bitwise_operations = {
    "RotateLeft",
    "RotateRight",
    "LShR",
    "Reverse",
}

backend_boolean_operations = {"And", "Or", "Not"}

backend_bitmod_operations = {"Concat", "Extract", "SignExt", "ZeroExt"}

backend_creation_operations = {"BoolV", "BVV", "FPV", "StringV"}

backend_symbol_creation_operations = {"BoolS", "BVS", "FPS", "StringS"}

backend_other_operations = {"If"}

backend_arithmetic_operations = {"SDiv", "SMod"}

backend_operations = (
    backend_comparator_operations
    | backend_bitwise_operations
    | backend_boolean_operations
    | backend_bitmod_operations
    | backend_creation_operations
    | backend_other_operations
    | backend_arithmetic_operations
)
backend_operations_vsa_compliant = (
    backend_bitwise_operations | backend_comparator_operations | backend_boolean_operations | backend_bitmod_operations
)
backend_operations_all = backend_operations | backend_operations_vsa_compliant

backend_fp_cmp_operations = {
    "fpLT",
    "fpLEQ",
    "fpGT",
    "fpGEQ",
    "fpEQ",
    "fpNEQ",
}

backend_fp_operations = {
    "FPS",
    "fpToFP",
    "fpToFPUnsigned",
    "fpToIEEEBV",
    "fpFP",
    "fpToSBV",
    "fpToUBV",
    "fpNeg",
    "fpSub",
    "fpAdd",
    "fpMul",
    "fpDiv",
    "fpAbs",
    "fpIsNaN",
    "fpIsInf",
    "fpSqrt",
} | backend_fp_cmp_operations

backend_strings_operations = {
    "StrSubstr",
    "StrReplace",
    "StrConcat",
    "StrLen",
    "StrContains",
    "StrPrefixOf",
    "StrSuffixOf",
    "StrIndexOf",
    "StrToInt",
    "StrIsDigit",
    "IntToStr",
}

opposites = {
    "__add__": "__radd__",
    "__radd__": "__add__",
    "__truediv__": "__rtruediv__",
    "__rtruediv__": "__truediv__",
    "__floordiv__": "__rfloordiv__",
    "__rfloordiv__": "__floordiv__",
    "__mul__": "__rmul__",
    "__rmul__": "__mul__",
    "__sub__": "__rsub__",
    "__rsub__": "__sub__",
    "__mod__": "__rmod__",
    "__rmod__": "__mod__",
    "__eq__": "__eq__",
    "__ne__": "__ne__",
    "ULT": "UGT",
    "UGT": "ULT",
    "ULE": "UGE",
    "UGE": "ULE",
    "SLT": "SGT",
    "SGT": "SLT",
    "SLE": "SGE",
    "SGE": "SLE",
    # '__neg__':
    # '__abs__':
    # '__invert__':
    "__or__": "__ror__",
    "__ror__": "__or__",
    "__and__": "__rand__",
    "__rand__": "__and__",
    "__xor__": "__rxor__",
    "__rxor__": "__xor__",
    "__lshift__": "__rlshift__",
    "__rlshift__": "__lshift__",
    "__rshift__": "__rrshift__",
    "__rrshift__": "__rshift__",
}

reversed_ops = {
    "__radd__": "__add__",
    "__rand__": "__and__",
    "__rfloordiv__": "__floordiv__",
    "__rlshift__": "__lshift__",
    "__rmod__": "__mod__",
    "__rmul__": "__mul__",
    "__ror__": "__or__",
    "__rrshift__": "__rshift__",
    "__rsub__": "__sub__",
    "__rtruediv__": "__truediv__",
    "__rxor__": "__xor__",
}

inverse_operations = {
    "__eq__": "__ne__",
    "__ne__": "__eq__",
    "ULT": "UGE",
    "UGE": "ULT",
    "UGT": "ULE",
    "ULE": "UGT",
    "SLT": "SGE",
    "SGE": "SLT",
    "SLE": "SGT",
    "SGT": "SLE",
}

leaf_operations = backend_symbol_creation_operations | backend_creation_operations
leaf_operations_concrete = backend_creation_operations
leaf_operations_symbolic = backend_symbol_creation_operations
leaf_operations_symbolic_with_union = leaf_operations_symbolic | {"union"}

#
# Reversibility
#

not_invertible = {"union"}
reverse_distributable = {
    "widen",
    "union",
    "intersection",
    "__invert__",
    "__or__",
    "__ror__",
    "__and__",
    "__rand__",
    "__xor__",
    "__rxor__",
}

infix = {
    "__add__": "+",
    "__sub__": "-",
    "__mul__": "*",
    "__floordiv__": "/",
    "__truediv__": "/",  # the raw / operator should use integral semantics on bitvectors
    "__mod__": "%",
    "__eq__": "==",
    "__ne__": "!=",
    "UGE": ">=",
    "ULE": "<=",
    "UGT": ">",
    "ULT": "<",
    "SGE": ">=s",
    "SLE": "<=s",
    "SGT": ">s",
    "SLT": "<s",
    "SDiv": "/s",
    "SMod": "%s",
    "__or__": "|",
    "__and__": "&",
    "__xor__": "^",
    "__lshift__": "<<",
    "__rshift__": ">>",
    "And": "&&",
    "Or": "||",
    "Concat": "..",
}

prefix = {
    "Not": "!",
    "__neg__": "-",
    "__invert__": "~",
}

op_precedence = {  # based on https://en.cppreference.com/w/c/language/operator_precedence
    # precedence: 2
    "Not": 2,
    "__neg__": 2,
    "__invert__": 2,
    # precedence: 3
    "__mul__": 3,
    "__floordiv__": 3,
    "__truediv__": 3,  # the raw / operator should use integral semantics on bitvectors
    "__mod__": 3,
    "SDiv": 3,
    "SMod": 3,
    # precedence: 4
    "__add__": 4,
    "__sub__": 4,
    # precedence: 5
    "__lshift__": 5,
    "__rshift__": 5,
    # precedence: 6
    "__ge__": 6,
    "__le__": 6,
    "__gt__": 6,
    "__lt__": 6,
    "UGE": 6,
    "ULE": 6,
    "UGT": 6,
    "ULT": 6,
    "SGE": 6,
    "SLE": 6,
    "SGT": 6,
    "SLT": 6,
    # precedence: 7
    "__eq__": 7,
    "__ne__": 7,
    # precedence: 8
    "__and__": 8,
    # precedence: 9
    "__xor__": 9,
    # precedence: 10
    "__or__": 10,
    # precedence: 11
    "And": 11,
    # precedence: 12
    "Or": 12,
    # 'Concat': '..',
}

commutative_operations = {
    "__and__",
    "__or__",
    "__xor__",
    "__add__",
    "__mul__",
    "And",
    "Or",
    "Xor",
}

bound_ops = {
    "Not": "__invert__",
    "And": "__and__",
    "Or": "__or__",
}
